import json
import pathlib
import typing

import pandas
import pydantic
import pynwb
import typing_extensions

from .._inspection._inspection_result import InspectionResult
from ..bids_models._base_metadata_model import BaseMetadataContainerModel, BaseMetadataModel


class Channel(BaseMetadataModel):
    channel_id: str
    electrode_id: str
    type: typing.Literal["EXT"] = "EXT"  # TODO
    unit: typing.Literal["V"] = "V"
    sampling_frequency: float | None = None
    gain: float | None = None


class ChannelTable(BaseMetadataContainerModel):
    channels: list[Channel]

    @pydantic.computed_field
    @property
    def messages(self) -> list[InspectionResult]:
        """
        All messages from contained session converters.

        These can accumulate over time based on which instance methods have been called.
        """
        messages = [message for channel in self.channels for message in channel.messages]
        messages.sort(key=lambda message: (-message.category.value, -message.severity.value, message.title))
        return messages

    @classmethod
    @pydantic.validate_call
    def from_nwbfiles(cls, nwbfiles: list[pydantic.InstanceOf[pynwb.NWBFile]]) -> typing_extensions.Self | None:
        if len(nwbfiles) > 1:
            message = "Conversion of multiple NWB files per session is not yet supported."
            raise NotImplementedError(message)
        nwbfile = nwbfiles[0]

        if nwbfile.electrodes is None:
            return None

        if len(nwbfile.electrodes) == 0:
            message = (
                "The NWB file contains an electrodes table, but it is empty. "
                "Please raise an issue on https://github.com/con/nwb2bids/issues to discuss."
            )
            raise NotImplementedError(message)

        # Only scan electrical series listed under acquisition since those under processing can downsample the rate
        raw_electrical_series = [
            neurodata_object
            for neurodata_object in nwbfile.acquisition
            if isinstance(neurodata_object, pynwb.ecephys.ElectricalSeries)
        ]
        if len(raw_electrical_series) > 1:
            # TODO: form a map of electrode to rate/gate based on ElectricalSeries linkage
            message = (
                "Support for automatic extraction of rates/gains from multiple ElectricalSeries is not yet implemented."
            )
            raise NotImplementedError(message)

        sampling_frequency = None
        gain = None
        if len(raw_electrical_series) == 1:
            electrical_series = raw_electrical_series[0]
            if electrical_series.rate is None:
                message = (
                    "Support for automatic extraction of rate from ElectricalSeries with "
                    "timestamps is not yet implemented."
                )
                raise NotImplementedError(message)

            sampling_frequency = electrical_series.rate
            gain = electrical_series.conversion

        channels = [
            Channel(
                channel_id=(
                    str(channel_name.values[0])
                    if (channel_name := electrode.get("channel_name", None)) is not None
                    else str(electrode.index[0])
                ),
                electrode_id=(
                    str(contact_ids.values[0])
                    if (contact_ids := electrode.get("contact_ids", None)) is not None
                    else str(electrode.index[0])
                ),
                type="EXT",
                unit="V",
                sampling_frequency=sampling_frequency,
                gain=gain,
            )
            for electrode in nwbfile.electrodes
        ]
        return cls(channels=channels)

    @pydantic.validate_call
    def to_tsv(self, file_path: str | pathlib.Path):
        """
        Write the channels data to a TSV file.

        Parameters
        ----------
        file_path : path
            The path where the TSV file will be saved.
        """
        data = []
        for channel in self.channels:
            model_dump = channel.model_dump()
            data.append(model_dump)

        data_frame = pandas.DataFrame(data=data)
        data_frame.to_csv(path_or_buf=file_path, sep="\t", index=False)

    @pydantic.validate_call
    def to_json(self, file_path: str | pathlib.Path) -> None:
        """
        Save the channels information to a JSON file.

        Parameters
        ----------
        file_path : path
            The path to the output JSON file.
        """
        with file_path.open(mode="w") as file_stream:
            json.dump(obj=dict(), fp=file_stream, indent=4)



import collections
import pathlib
import shutil
import typing

import pydantic
import typing_extensions

from .._converters._base_converter import BaseConverter
from .._inspection._inspection_result import InspectionResult
from .._tools import cache_read_nwb
from ..bids_models import BidsSessionMetadata


class SessionConverter(BaseConverter):
    """
    Initialize a converter of NWB files to BIDS format.
    """

    session_id: str = pydantic.Field(
        description="A unique session identifier.",
    )
    nwbfile_paths: list[pydantic.FilePath] | list[pydantic.HttpUrl] = pydantic.Field(
        description="List of file paths or URLs of NWB files which share this session ID.", min_length=1
    )
    session_metadata: BidsSessionMetadata | None = pydantic.Field(
        description="BIDS metadata extracted for this session.", default=None
    )
    messages: list[InspectionResult] = pydantic.Field(
        description="List of auto-detected suggestions.", ge=0, default_factory=list
    )

    @classmethod
    @pydantic.validate_call
    def from_nwb_paths(
        cls,
        nwb_paths: list[pydantic.FilePath | pydantic.DirectoryPath] = pydantic.Field(min_length=1),
    ) -> list[typing_extensions.Self]:
        """
        Initialize a list of session converters from a list of NWB file paths.

        Automatically associates multiple NWB files to a single session according to session ID.

        Parameters
        ----------
        nwb_paths : iterable of file and directory paths
            An iterable of NWB file paths and directories containing NWB files.

        Returns
        -------
        A list of SessionConverter instances, one per unique session ID.
        """
        all_nwbfile_paths = []
        for nwb_path in nwb_paths:
            if nwb_path.is_file():
                all_nwbfile_paths.append(nwb_path)
            elif nwb_path.is_dir():
                all_nwbfile_paths += list(nwb_path.rglob(pattern="*.nwb"))

        unique_session_id_to_nwbfile_paths = collections.defaultdict(list)
        for nwbfile_path in all_nwbfile_paths:
            unique_session_id_to_nwbfile_paths[cache_read_nwb(nwbfile_path).session_id].append(nwbfile_path)

        session_converters = [
            cls(session_id=session_id, nwbfile_paths=nwbfile_paths)
            for session_id, nwbfile_paths in unique_session_id_to_nwbfile_paths.items()
        ]
        return session_converters

    def extract_metadata(self) -> None:
        if self.session_metadata is None:
            self.session_metadata = BidsSessionMetadata.from_nwbfile_paths(nwbfile_paths=self.nwbfile_paths)
            self.messages += self.session_metadata.messages

    @pydantic.validate_call
    def convert_to_bids_session(
        self,
        bids_directory: str | pathlib.Path | None = None,
        file_mode: typing.Literal["move", "copy", "symlink", "auto"] = "auto",
    ) -> None:
        """
        Convert the NWB file to a BIDS session directory.

        Parameters
        ----------
        bids_directory : directory path
            The path to the directory where the BIDS dataset will be created.
        file_mode : one of "move", "copy", "symlink", or "auto", default: "auto"
            Specifies how to handle the NWB files when converting to BIDS format.
            - "move": Move the files to the BIDS directory.
            - "copy": Copy the files to the BIDS directory.
            - "symlink": Create symbolic links to the files in the BIDS directory.
            - "auto": Decides between all the above based on the system, with preference for linking when possible.
        """
        if len(self.nwbfile_paths) > 1:
            message = (
                "Conversion of multiple NWB files per session is not yet supported. "
                "Please raise an issue on https://github.com/con/nwb2bids/issues/new to discuss the use case."
            )
            raise NotImplementedError(message)
        bids_directory = self._handle_bids_directory(bids_directory=bids_directory)
        file_mode = self._handle_file_mode(file_mode=file_mode)

        if self.session_metadata is None:
            self.extract_metadata()

        subject_id = self.session_metadata.participant.participant_id
        session_subdirectory = bids_directory / f"sub-{subject_id}" / f"ses-{self.session_id}" / "ecephys"
        session_subdirectory.mkdir(parents=True, exist_ok=True)

        self.write_ecephys_files(bids_directory=bids_directory)
        if self.session_metadata.events is not None:
            self.write_events_files(bids_directory=bids_directory)

        # TODO: determine icephys or ecephys suffix
        for nwbfile_path in self.nwbfile_paths:
            session_file_path = session_subdirectory / f"sub-{subject_id}_ses-{self.session_id}_ecephys.nwb"

            # If not a local path, then it is a URL, so write simple 'symlink' pointing to the URL
            if not isinstance(nwbfile_path, pathlib.Path):
                with session_file_path.open(mode="w") as file_stream:
                    file_stream.write(str(nwbfile_path))
                continue

            if file_mode == "copy":
                shutil.copy(src=nwbfile_path, dst=session_file_path)
            elif file_mode == "move":
                shutil.move(src=nwbfile_path, dst=session_file_path)
            elif file_mode == "symlink":
                session_file_path.symlink_to(target=nwbfile_path)

    @pydantic.validate_call
    def write_ecephys_files(self, bids_directory: str | pathlib.Path | None = None) -> None:
        """
        Write the `_probes`, `_channels`, and `_electrodes` metadata files, both `.tsv` and `.json`, for this session .

        Parameters
        ----------
        bids_directory : directory path
            The path to the directory where the BIDS dataset will be created.
        """
        if len(self.nwbfile_paths) > 1:
            message = "Conversion of multiple NWB files per session is not yet supported."
            raise NotImplementedError(message)

        if (
            self.session_metadata.probe_table is None
            and self.session_metadata.channel_table is None
            and self.session_metadata.electrode_table is None
        ):
            return
        bids_directory = self._handle_bids_directory(bids_directory=bids_directory)
        ecephys_directory = self._establish_ecephys_subdirectory(bids_directory=bids_directory)

        file_prefix = f"sub-{self.session_metadata.participant.participant_id}_ses-{self.session_id}"

        if self.session_metadata.probe_table is not None:
            probes_tsv_file_path = ecephys_directory / f"{file_prefix}_probes.tsv"
            self.session_metadata.probe_table.to_tsv(file_path=probes_tsv_file_path)

            probes_json_file_path = ecephys_directory / f"{file_prefix}_probes.json"
            self.session_metadata.probe_table.to_json(file_path=probes_json_file_path)

        if self.session_metadata.channel_table is not None:
            channels_tsv_file_path = ecephys_directory / f"{file_prefix}_channels.tsv"
            self.session_metadata.channel_table.to_tsv(file_path=channels_tsv_file_path)

            channels_json_file_path = ecephys_directory / f"{file_prefix}_channels.json"
            self.session_metadata.channel_table.to_json(file_path=channels_json_file_path)

        if self.session_metadata.electrode_table is not None:
            electrodes_tsv_file_path = ecephys_directory / f"{file_prefix}_electrodes.tsv"
            self.session_metadata.electrode_table.to_tsv(file_path=electrodes_tsv_file_path)

            electrodes_json_file_path = ecephys_directory / f"{file_prefix}_electrodes.json"
            self.session_metadata.electrode_table.to_json(file_path=electrodes_json_file_path)

    @pydantic.validate_call
    def write_events_files(self, bids_directory: str | pathlib.Path | None = None) -> None:
        """
        Write the `_events.tsv` and `_events.json` files for this session.

        Parameters
        ----------
        bids_directory : directory path
            The path to the directory where the BIDS dataset will be created.
        """
        if self.session_metadata.events is None:
            message = "No events metadata found in the session metadata - unable to write events TSV."
            raise ValueError(message)
        if len(self.nwbfile_paths) > 1:
            message = "Conversion of multiple NWB files per session is not yet supported."
            raise NotImplementedError(message)
        bids_directory = self._handle_bids_directory(bids_directory=bids_directory)
        ecephys_directory = self._establish_ecephys_subdirectory(bids_directory=bids_directory)

        file_prefix = f"sub-{self.session_metadata.participant.participant_id}_ses-{self.session_id}"
        session_events_tsv_file_path = ecephys_directory / f"{file_prefix}_events.tsv"
        self.session_metadata.events.to_tsv(file_path=session_events_tsv_file_path)

        # TODO: add merging from top-level additional metadata mechanism
        session_events_metadata_file_path = ecephys_directory / f"{file_prefix}_events.json"
        self.session_metadata.events.to_json(file_path=session_events_metadata_file_path)

        # IDEA: check events.json files, see if all are the same and if so remove the duplicates and move to outer level

    def _establish_ecephys_subdirectory(self, bids_directory: pathlib.Path) -> pathlib.Path:
        subject_directory = bids_directory / f"sub-{self.session_metadata.participant.participant_id}"
        subject_directory.mkdir(exist_ok=True)
        session_directory = subject_directory / f"ses-{self.session_id}"
        session_directory.mkdir(exist_ok=True)
        ecephys_directory = session_directory / "ecephys"
        ecephys_directory.mkdir(exist_ok=True)

        return ecephys_directory



import pathlib
import typing

import pydantic

from .._converters._dataset_converter import DatasetConverter
from .._inspection._inspection_result import InspectionResult


@pydantic.validate_call
def convert_nwb_dataset(
    *,
    nwb_paths: list[pydantic.FilePath | pydantic.DirectoryPath] = pydantic.Field(min_length=1),
    bids_directory: str | pathlib.Path | None = None,
    file_mode: typing.Literal["move", "copy", "symlink", "auto"] = "auto",
    additional_metadata_file_path: pydantic.FilePath | None = None,
) -> list[InspectionResult]:
    """
    Convert a dataset of NWB files to a BIDS dataset.

    Parameters
    ----------
    nwb_paths : any iterable of file or directory paths
        An iterable of NWB file paths and directories containing NWB files.
    bids_directory : string or path, optional
        The path to the directory where the BIDS dataset will be created.
        If None, uses the current working directory.
    file_mode : one of "move", "copy", "symlink", or "auto", default "auto"
        The file operation to use when adding files to the BIDS dataset.
        "move": Move the files (deletes the original files).
        "copy": Copy the files (original files remain unchanged).
        "symlink": Create symbolic links to the original files.
        "auto": Use "symlink" if the source and destination are on the same filesystem, otherwise use "copy".
    additional_metadata_file_path : file path, optional
        The path to a YAML file containing additional metadata not included within the NWB files
        that you wish to include in the BIDS dataset.

    Returns
    -------
    notifications : list of InspectionResult or None
        A list of inspection results, or None if there are no internal messages.
    """
    converter = DatasetConverter.from_nwb_paths(
        nwb_paths=nwb_paths,
        additional_metadata_file_path=additional_metadata_file_path,
    )
    converter.extract_metadata()
    converter.convert_to_bids_dataset(bids_directory=bids_directory, file_mode=file_mode)

    return converter.messages



import json
import pathlib

import pytest

import nwb2bids


@pytest.mark.parametrize(
    "test_case",
    [
        "explicit_empty",
        "explicit_exists_valid_bids",
        "implicit_empty",
        "implicit_exists_valid_bids",
    ],
)
def test_allowed_directory_conditions(
    test_case: str,
    minimal_nwbfile_path: pathlib.Path,
    additional_metadata_file_path: pathlib.Path,
    temporary_bids_directory: pathlib.Path,
    monkeypatch,
):
    dataset_description_file_path = temporary_bids_directory / "dataset_description.json"

    if "valid" in test_case:
        dataset_description_file_path.write_text(data='{"BIDSVersion": "1.10"}')

    bids_directory = None
    if "implicit" in test_case:
        monkeypatch.chdir(temporary_bids_directory)
    else:
        bids_directory = temporary_bids_directory

    nwb_paths = [minimal_nwbfile_path]
    dataset_converter = nwb2bids.DatasetConverter.from_nwb_paths(
        nwb_paths=nwb_paths, additional_metadata_file_path=additional_metadata_file_path
    )
    dataset_converter.extract_metadata()
    dataset_converter.write_dataset_description(bids_directory=bids_directory)

    expected_structure = {temporary_bids_directory: {"directories": set(), "files": {"dataset_description.json"}}}
    nwb2bids.testing.assert_subdirectory_structure(
        directory=temporary_bids_directory, expected_structure=expected_structure
    )

    dataset_description_file_path = temporary_bids_directory / "dataset_description.json"
    with dataset_description_file_path.open(mode="r") as file_stream:
        dataset_description_json = json.load(fp=file_stream)

    expected_dataset_description = {
        "Name": "test",
        "Description": "TODO",
        "BIDSVersion": "1.10",
        "DatasetType": "raw",
        "License": "CC-BY-4.0",
        "Authors": ["Cody Baker", "Yaroslav Halchenko"],
    }
    assert dataset_description_json == expected_dataset_description


@pytest.mark.parametrize(
    "test_case",
    [
        "explicit_no_dataset_description",
        "explicit_no_bids_version",
        "implicit_no_dataset_description",
        "implicit_no_bids_version",
    ],
)
def test_disallowed_directory_conditions(
    test_case: str,
    minimal_nwbfile_path: pathlib.Path,
    additional_metadata_file_path: pathlib.Path,
    temporary_bids_directory: pathlib.Path,
    monkeypatch,
):
    dataset_description_file_path = temporary_bids_directory / "dataset_description.json"

    bids_directory = None
    if "implicit" in test_case:
        monkeypatch.chdir(temporary_bids_directory)
    else:
        bids_directory = temporary_bids_directory

    if "no_dataset_description" in test_case:
        other_file_path = dataset_description_file_path.parent / "other_file.txt"
        other_file_path.write_text(data="This is not a BIDS dataset.")

        expected_exception = "missing 'dataset_description.json'"
    elif "no_bids_version" in test_case:
        dataset_description_file_path.write_text(data='{"Name": "test"}')

        expected_exception = "missing 'BIDSVersion' in 'dataset_description.json'"

    with pytest.raises(expected_exception=ValueError, match=expected_exception):
        nwb_paths = [minimal_nwbfile_path]
        dataset_converter = nwb2bids.DatasetConverter.from_nwb_paths(
            nwb_paths=nwb_paths, additional_metadata_file_path=additional_metadata_file_path
        )
        dataset_converter.extract_metadata()
        dataset_converter.write_dataset_description(bids_directory=bids_directory)



"""Configuration file for the Sphinx documentation builder."""

import sys
from pathlib import Path

# Add source directory to path for autodoc
sys.path.insert(0, str(Path(__file__).resolve().parents[2] / "src"))

project = "nwb2bids"
copyright = "2025, Cody Baker"
author = "Cody Baker"

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.napoleon",
    "sphinx.ext.viewcode",
]

exclude_patterns = []

html_theme = "pydata_sphinx_theme"

html_theme_options = {
    "github_url": "https://github.com/con/nwb2bids",
    "use_edit_page_button": True,
    "show_toc_level": 2,
}

html_context = {
    "github_user": "con",
    "github_repo": "nwb2bids",
    "github_version": "main",
    "doc_path": "docs/source",
}

# Napoleon settings
napoleon_google_docstring = False
napoleon_numpy_docstring = True
napoleon_use_param = False
napoleon_use_ivar = True

# Autodoc settings
autoclass_content = "both"
autodoc_default_options = {
    "members": True,
    "member-order": "bysource",
    "private-members": False,
    "undoc-members": True,
    "show-inheritance": False,
    "exclude-members": "__new__",
}
add_module_names = False
toc_object_entries_show_parents = "hide"

# Format signatures for better readability
autodoc_typehints = "signature"
autodoc_typehints_format = "short"
python_maximum_signature_line_length = 88



import datetime
import json
import pathlib

import py.path
import pynwb
import pynwb.file
import pynwb.testing.mock.ecephys
import pynwb.testing.mock.file
import pytest

import nwb2bids

# TODO: add ndx-events
# TODO: add DynamicTable's in acquisition with *_time columns


def _make_minimal_nwbfile(session_id: str = "456") -> pynwb.NWBFile:
    """Creates a minimal NWB file for testing purposes."""
    nwbfile = pynwb.testing.mock.file.mock_NWBFile(session_id=session_id)

    subject = pynwb.file.Subject(
        subject_id="123",
        species="Mus musculus",
        sex="M",
    )
    nwbfile.subject = subject

    return nwbfile


@pytest.fixture()
def temporary_bids_directory(tmpdir: py.path.local) -> pathlib.Path:
    """Creates a temporary BIDS directory for testing purposes."""
    return pathlib.Path(tmpdir)


@pytest.fixture(scope="session")
def testing_files_directory(tmp_path_factory: pytest.TempPathFactory) -> pathlib.Path:
    """
    Creates a common temporary directory for all testing files.

    Testing files may be nested into subdirectories to test localization logic.
    """
    testing_files_directory = tmp_path_factory.mktemp(basename="nwb2bids_testing_files", numbered=False)
    return testing_files_directory


@pytest.fixture(scope="session")
def additional_metadata_file_path(testing_files_directory: pathlib.Path) -> pathlib.Path:
    additional_metadata_dictionary = {
        "dataset_description": {
            "Name": "test",
            "Description": "TODO",
            "BIDSVersion": "1.10",
            "DatasetType": "raw",
            "License": "CC-BY-4.0",
            "Authors": ["Cody Baker", "Yaroslav Halchenko"],
        }
    }

    additional_metadata_subdirectory = testing_files_directory / "additional_metadata"
    additional_metadata_subdirectory.mkdir(exist_ok=True)
    file_path = additional_metadata_subdirectory / "test_additional_metadata.json"
    with file_path.open(mode="w") as file_stream:
        json.dump(obj=additional_metadata_dictionary, fp=file_stream)

    return file_path


@pytest.fixture(scope="session")
def minimal_nwbfile_path(testing_files_directory: pathlib.Path) -> pathlib.Path:
    """
    A minimally valid NWB file for testing purposes.

    Does not contain any additional metadata beyond what is required for initialization of converter and metadata
    classes.
    """
    nwbfile = _make_minimal_nwbfile()

    minimal_subdirectory = testing_files_directory / "minimal"
    minimal_subdirectory.mkdir(exist_ok=True)
    nwbfile_path = minimal_subdirectory / "minimal.nwb"
    with pynwb.NWBHDF5IO(path=nwbfile_path, mode="w") as file_stream:
        file_stream.write(nwbfile)

    return nwbfile_path


@pytest.fixture(scope="session")
def ecephys_nwbfile_path(testing_files_directory: pathlib.Path) -> pathlib.Path:
    nwbfile = _make_minimal_nwbfile(session_id="789")

    pynwb.testing.mock.ecephys.mock_ElectricalSeries(name="test_electrical_series", nwbfile=nwbfile)

    ecephys_subdirectory = testing_files_directory / "ecephys"
    ecephys_subdirectory.mkdir(exist_ok=True)
    nwbfile_path = ecephys_subdirectory / "ecephys.nwb"
    with pynwb.NWBHDF5IO(path=nwbfile_path, mode="w") as file_stream:
        file_stream.write(nwbfile)

    return nwbfile_path


@pytest.fixture(scope="session")
def trials_events_nwbfile_path(testing_files_directory: pathlib.Path) -> pathlib.Path:
    nwbfile = _make_minimal_nwbfile()

    trials = nwb2bids.testing.mock_trials_table()
    nwbfile.trials = trials

    events_subdirectory = testing_files_directory / "trials"
    events_subdirectory.mkdir(exist_ok=True)
    nwbfile_path = events_subdirectory / "trials.nwb"
    with pynwb.NWBHDF5IO(path=nwbfile_path, mode="w") as file_stream:
        file_stream.write(nwbfile)

    return nwbfile_path


@pytest.fixture(scope="session")
def epochs_events_nwbfile_path(testing_files_directory: pathlib.Path) -> pathlib.Path:
    nwbfile = _make_minimal_nwbfile()

    epochs = nwb2bids.testing.mock_epochs_table()
    nwbfile.epochs = epochs

    events_subdirectory = testing_files_directory / "epochs"
    events_subdirectory.mkdir(exist_ok=True)
    nwbfile_path = events_subdirectory / "epochs.nwb"
    with pynwb.NWBHDF5IO(path=nwbfile_path, mode="w") as file_stream:
        file_stream.write(nwbfile)

    return nwbfile_path


@pytest.fixture(scope="session")
def multiple_events_nwbfile_path(testing_files_directory: pathlib.Path) -> pathlib.Path:
    nwbfile = _make_minimal_nwbfile()

    trials = nwb2bids.testing.mock_trials_table()
    nwbfile.trials = trials

    epochs = nwb2bids.testing.mock_epochs_table()
    nwbfile.epochs = epochs

    time_intervals = nwb2bids.testing.mock_time_intervals()
    nwbfile.add_acquisition(time_intervals)

    events_subdirectory = testing_files_directory / "multiple_events"
    events_subdirectory.mkdir(exist_ok=True)
    nwbfile_path = events_subdirectory / "multiple_events.nwb"
    with pynwb.NWBHDF5IO(path=nwbfile_path, mode="w") as file_stream:
        file_stream.write(nwbfile)

    return nwbfile_path


@pytest.fixture(scope="session")
def nwbfile_path_with_missing_session_id(testing_files_directory: pathlib.Path) -> pathlib.Path:
    nwbfile = pynwb.testing.mock.file.mock_NWBFile(session_id=None)

    subject = pynwb.file.Subject(
        subject_id="123",
        species="Mus musculus",
        sex="M",
    )
    nwbfile.subject = subject

    events_subdirectory = testing_files_directory / "missing_session_id"
    events_subdirectory.mkdir(exist_ok=True)
    nwbfile_path = events_subdirectory / "missing_session_id.nwb"
    with pynwb.NWBHDF5IO(path=nwbfile_path, mode="w") as file_stream:
        file_stream.write(nwbfile)

    return nwbfile_path


@pytest.fixture(scope="session")
def directory_with_multiple_nwbfiles(testing_files_directory: pathlib.Path) -> pathlib.Path:
    multiple_nwbfiles_subdirectory = testing_files_directory / "multiple_nwbfiles"
    multiple_nwbfiles_subdirectory.mkdir(exist_ok=True)

    for session_index in range(2):
        nwbfile = _make_minimal_nwbfile(session_id=f"session-{session_index}")

        nwbfile_path = multiple_nwbfiles_subdirectory / f"session_{session_index}.nwb"
        with pynwb.NWBHDF5IO(path=nwbfile_path, mode="w") as file_stream:
            file_stream.write(nwbfile)

    return multiple_nwbfiles_subdirectory


@pytest.fixture(scope="session")
def problematic_nwbfile_path_1(testing_files_directory: pathlib.Path) -> pathlib.Path:
    """
    An NWB file with 'problematic' metadata for testing the messaging feature.
    """
    nwbfile = pynwb.NWBFile(
        identifier="not a UUID",
        session_id="problematic1",
        session_description="",
        session_start_time=datetime.datetime.now().astimezone(),
    )

    subject = pynwb.file.Subject(
        subject_id="bad_subject_id",
        species="not a latin species, nor a ncbi taxon link or id",
        sex="bad sex specifier",
    )
    nwbfile.subject = subject

    problematic_subdirectory = testing_files_directory / "problematic"
    problematic_subdirectory.mkdir(exist_ok=True)
    nwbfile_path = problematic_subdirectory / "problematic1.nwb"
    with pynwb.NWBHDF5IO(path=nwbfile_path, mode="w") as file_stream:
        file_stream.write(nwbfile)

    return nwbfile_path


@pytest.fixture(scope="session")
def problematic_nwbfile_path_2(testing_files_directory: pathlib.Path) -> pathlib.Path:
    """
    A second NWB file with 'problematic' metadata for testing the messaging feature.
    """
    nwbfile = pynwb.NWBFile(
        identifier="not a UUID",
        session_id="problematic2",
        session_description="",
        session_start_time=datetime.datetime.now().astimezone(),
    )

    subject = pynwb.file.Subject(
        subject_id="bad subject id",
    )
    nwbfile.subject = subject

    problematic_subdirectory = testing_files_directory / "problematic"
    problematic_subdirectory.mkdir(exist_ok=True)
    nwbfile_path = problematic_subdirectory / "problematic2.nwb"
    with pynwb.NWBHDF5IO(path=nwbfile_path, mode="w") as file_stream:
        file_stream.write(nwbfile)

    return nwbfile_path


@pytest.fixture(scope="session")
def problematic_nwbfile_path_3(testing_files_directory: pathlib.Path) -> pathlib.Path:
    """
    A third NWB file with 'problematic' metadata (entirely missing entities) for testing the messaging feature.
    """
    nwbfile = pynwb.NWBFile(
        identifier="not a UUID",
        session_id="problematic3",
        session_description="",
        session_start_time=datetime.datetime.now().astimezone(),
    )

    problematic_subdirectory = testing_files_directory / "problematic"
    problematic_subdirectory.mkdir(exist_ok=True)
    nwbfile_path = problematic_subdirectory / "problematic3.nwb"
    with pynwb.NWBHDF5IO(path=nwbfile_path, mode="w") as file_stream:
        file_stream.write(nwbfile)

    return nwbfile_path



import pathlib

import pytest

import nwb2bids


@pytest.mark.remote
def test_remote_convert_nwb_dataset(temporary_bids_directory: pathlib.Path):
    dataset_converter = nwb2bids.DatasetConverter.from_remote_dandiset(dandiset_id="000003", limit=2)
    dataset_converter.extract_metadata()
    dataset_converter.convert_to_bids_dataset(bids_directory=temporary_bids_directory)

    expected_structure = {
        temporary_bids_directory: {
            "directories": {"sub-YutaMouse20"},
            "files": {"dataset_description.json", "participants.json", "participants.tsv"},
        },
        temporary_bids_directory
        / "sub-YutaMouse20": {
            "directories": {"ses-YutaMouse20-140321", "ses-YutaMouse20-140327"},
            "files": {"sub-YutaMouse20_sessions.json", "sub-YutaMouse20_sessions.tsv"},
        },
        temporary_bids_directory
        / "sub-YutaMouse20"
        / "ses-YutaMouse20-140321": {
            "directories": {"ecephys"},
            "files": set(),
        },
        temporary_bids_directory
        / "sub-YutaMouse20"
        / "ses-YutaMouse20-140327": {
            "directories": {"ecephys"},
            "files": set(),
        },
        temporary_bids_directory
        / "sub-YutaMouse20"
        / "ses-YutaMouse20-140321"
        / "ecephys": {
            "directories": set(),
            "files": {
                "sub-YutaMouse20_ses-YutaMouse20-140321_channels.json",
                "sub-YutaMouse20_ses-YutaMouse20-140321_channels.tsv",
                "sub-YutaMouse20_ses-YutaMouse20-140321_ecephys.nwb",
                "sub-YutaMouse20_ses-YutaMouse20-140321_electrodes.json",
                "sub-YutaMouse20_ses-YutaMouse20-140321_electrodes.tsv",
                "sub-YutaMouse20_ses-YutaMouse20-140321_probes.json",
                "sub-YutaMouse20_ses-YutaMouse20-140321_probes.tsv",
            },
        },
        temporary_bids_directory
        / "sub-YutaMouse20"
        / "ses-YutaMouse20-140327"
        / "ecephys": {
            "directories": set(),
            "files": {
                "sub-YutaMouse20_ses-YutaMouse20-140327_channels.json",
                "sub-YutaMouse20_ses-YutaMouse20-140327_channels.tsv",
                "sub-YutaMouse20_ses-YutaMouse20-140327_ecephys.nwb",
                "sub-YutaMouse20_ses-YutaMouse20-140327_electrodes.json",
                "sub-YutaMouse20_ses-YutaMouse20-140327_electrodes.tsv",
                "sub-YutaMouse20_ses-YutaMouse20-140327_probes.json",
                "sub-YutaMouse20_ses-YutaMouse20-140327_probes.tsv",
            },
        },
    }
    nwb2bids.testing.assert_subdirectory_structure(
        directory=temporary_bids_directory, expected_structure=expected_structure
    )



# import json
# import typing

# import pydantic
# import pynwb
#
#
# class EcephysMetadata(pydantic.BaseModel):
#     """
#     General device and session metadata extracted from NWB files.
#
#     While NWB treats this information as high-level session-specific metadata, BIDS treats these fields as
#     modality specific and as pertaining to 'parameters' that can vary.
#
#     This should typically be written to two files:
#       - `acq-general_ecephys.json`
#       - `acq-
#     """
#
#     InstitutionName: str = pydantic.Field(
#         description="The name of the institution in charge of the equipment that produced the measurements."
#     )
#
#     model_config = pydantic.ConfigDict(
#         validate_assignment=True,  # Re-validate model on mutation
#         extra="allow",  # Allow additional custom fields
#     )
#
#     @classmethod
#     @pydantic.validate_call
#     def from_nwbfiles(cls, nwbfiles: list[pynwb.NWBFile]) -> typing_extensions.Self:
#         """
#         Extracts all unique general metadata from the in-memory NWBFile objects.
#         """
#         institution_names = {nwbfile.institution for nwbfile in nwbfiles if nwbfile.institution is not None}
#
#         general_metadata = cls(
#             InstitutionName=institution_names,
#         )
#         return subject
#
#     @pydantic.validate_call
#     def to_json(self, file_path: str | pathlib.Path) -> None:
#         """
#         Save the general metadata to a JSON file.
#
#         Parameters
#         ----------
#         file_path : str or pathlib.Path
#             The path to the file where the metadata will be saved.
#         """
#         with file_path.open(mode="w") as file_stream:
#             json.dump(obj=self.model_dump(), fp=file_stream, indent=4)



import json
import typing

import pydantic
import typing_extensions

from ._base_metadata_model import BaseMetadataModel


class DatasetDescription(BaseMetadataModel):
    """
    Schema for the dataset description in BIDS format.
    """

    Name: str = pydantic.Field(description="Name of the dataset.")
    BIDSVersion: str | None = pydantic.Field(
        description="The version of the BIDS standard that was used.",
        pattern=r"^\d+\.\d+(?:\.\d+)?$",
    )
    Description: str | None = pydantic.Field(
        description="Description of the dataset.",
        default=None,
    )
    DatasetType: typing.Literal["raw", "derivative"] = pydantic.Field(
        description="The interpretation of the dataset.",
        default="raw",
    )
    Authors: list[str] | None = pydantic.Field(
        description="List of individuals who contributed to the creation/curation of the dataset.",
        default=None,
    )
    License: typing.Literal["CC-BY-4.0", "CC0-1.0"] | None = pydantic.Field(
        description="License under which the dataset is released.",
        default=None,
    )

    @classmethod
    @pydantic.validate_call
    def from_file_path(cls, file_path: pydantic.FilePath) -> typing_extensions.Self | None:
        """
        Load the BIDS dataset description from a JSON file.

        Parameters
        ----------
        file_path : FilePath
            Path to the JSON file containing the dataset metadata.

        Returns
        -------
        BidsDatasetMetadata or None
            An instance of DatasetDescription populated with data from the file.
        """
        with file_path.open(mode="r") as file_stream:
            dictionary = json.load(fp=file_stream)

        dataset_description_content = dictionary.get("dataset_description", None)
        if dataset_description_content is None:
            return None

        dataset_metadata = cls(**dataset_description_content)
        return dataset_metadata



"""Unit tests for the `SessionConverter` class and its methods."""

import json
import pathlib

import pandas

import nwb2bids


def test_session_converter_directory_initialization(
    minimal_nwbfile_path: pathlib.Path, temporary_bids_directory: pathlib.Path
):
    nwb_paths = [minimal_nwbfile_path.parent]
    session_converters = nwb2bids.SessionConverter.from_nwb_paths(nwb_paths=nwb_paths)

    assert isinstance(session_converters, list)
    assert len(session_converters) == 1
    assert isinstance(session_converters[0], nwb2bids.SessionConverter)


def test_session_converter_file_path_initialization(
    minimal_nwbfile_path: pathlib.Path, temporary_bids_directory: pathlib.Path
):
    nwb_paths = [minimal_nwbfile_path]
    session_converters = nwb2bids.SessionConverter.from_nwb_paths(nwb_paths=nwb_paths)

    assert isinstance(session_converters, list)
    assert len(session_converters) == 1
    assert isinstance(session_converters[0], nwb2bids.SessionConverter)


def test_session_converter_both_file_and_directory_initialization(
    minimal_nwbfile_path: pathlib.Path, ecephys_nwbfile_path: pathlib.Path, temporary_bids_directory: pathlib.Path
):
    nwb_paths = [minimal_nwbfile_path.parent, ecephys_nwbfile_path]
    session_converters = nwb2bids.SessionConverter.from_nwb_paths(nwb_paths=nwb_paths)

    assert isinstance(session_converters, list)
    assert len(session_converters) == 2
    assert all(isinstance(converter, nwb2bids.SessionConverter) for converter in session_converters)


def test_session_converter_metadata_extraction(
    minimal_nwbfile_path: pathlib.Path, temporary_bids_directory: pathlib.Path
):
    nwb_paths = [minimal_nwbfile_path]
    session_converters = nwb2bids.SessionConverter.from_nwb_paths(nwb_paths=nwb_paths)
    session_converters[0].extract_metadata()

    expected_session_metadata = nwb2bids.bids_models.BidsSessionMetadata(
        session_id="456",
        participant=nwb2bids.bids_models.Participant(participant_id="123", species="Mus musculus", sex="M"),
    )
    assert session_converters[0].session_metadata == expected_session_metadata


def test_session_converter_write_ecephys_metadata(
    ecephys_nwbfile_path: pathlib.Path, temporary_bids_directory: pathlib.Path
):
    nwb_paths = [ecephys_nwbfile_path]
    session_converters = nwb2bids.SessionConverter.from_nwb_paths(nwb_paths=nwb_paths)
    session_converter = session_converters[0]
    session_converter.extract_metadata()
    session_converter.write_ecephys_files(bids_directory=temporary_bids_directory)

    expected_structure = {
        temporary_bids_directory: {"directories": {"sub-123"}, "files": {"dataset_description.json"}},
        temporary_bids_directory
        / "sub-123": {
            "directories": {"ses-789"},
            "files": set(),
        },
        temporary_bids_directory
        / "sub-123"
        / "ses-789": {
            "directories": {"ecephys"},
            "files": set(),
        },
        temporary_bids_directory
        / "sub-123"
        / "ses-789"
        / "ecephys": {
            "directories": set(),
            "files": {
                "sub-123_ses-789_probes.tsv",
                "sub-123_ses-789_probes.json",
                "sub-123_ses-789_channels.tsv",
                "sub-123_ses-789_channels.json",
                "sub-123_ses-789_electrodes.tsv",
                "sub-123_ses-789_electrodes.json",
            },
        },
    }
    nwb2bids.testing.assert_subdirectory_structure(
        directory=temporary_bids_directory, expected_structure=expected_structure
    )

    # TODO: assert contents of TSV and JSON files


def test_session_converter_write_events_metadata(
    trials_events_nwbfile_path: pathlib.Path, temporary_bids_directory: pathlib.Path
):
    nwb_paths = [trials_events_nwbfile_path]
    session_converters = nwb2bids.SessionConverter.from_nwb_paths(nwb_paths=nwb_paths)
    session_converter = session_converters[0]
    session_converter.extract_metadata()
    session_converter.write_events_files(bids_directory=temporary_bids_directory)

    expected_structure = {
        temporary_bids_directory: {"directories": {"sub-123"}, "files": {"dataset_description.json"}},
        temporary_bids_directory
        / "sub-123": {
            "directories": {"ses-456"},
            "files": set(),
        },
        temporary_bids_directory
        / "sub-123"
        / "ses-456": {
            "directories": {"ecephys"},
            "files": set(),
        },
        temporary_bids_directory
        / "sub-123"
        / "ses-456"
        / "ecephys": {
            "directories": set(),
            "files": {"sub-123_ses-456_events.tsv", "sub-123_ses-456_events.json"},
        },
    }
    nwb2bids.testing.assert_subdirectory_structure(
        directory=temporary_bids_directory, expected_structure=expected_structure
    )

    sessions_tsv_file_path = temporary_bids_directory / "sub-123" / "ses-456" / "ecephys" / "sub-123_ses-456_events.tsv"
    sessions_data_frame = pandas.read_csv(filepath_or_buffer=sessions_tsv_file_path, sep="\t")

    expected_data_frame = pandas.DataFrame(
        {
            "onset": [0.0, 2.0, 5.0, 5.5],
            "duration": [1.0, 1.0, 0.5, 0.5],
            "nwb_table": ["trials", "trials", "trials", "trials"],
            "trial_condition": ["A", "B", "C", "D"],
        },
    )
    pandas.testing.assert_frame_equal(left=sessions_data_frame, right=expected_data_frame)

    sessions_json_file_path = (
        temporary_bids_directory / "sub-123" / "ses-456" / "ecephys" / "sub-123_ses-456_events.json"
    )
    with sessions_json_file_path.open(mode="r") as file_stream:
        sessions_json = json.load(fp=file_stream)
    expected_sessions_json = {
        "trials": {"Description": "A mock trials table."},
        "nwb_table": {
            "nwb_table": {
                "Description": "The name of the NWB table from which this event was extracted.",
                "Levels": {"trials": "The 'trials' table in the NWB file."},
                "HED": {"trials": "Experimental-trial"},
            }
        },
    }
    assert sessions_json == expected_sessions_json



import typing

import pydantic

from .._inspection._inspection_result import Category, InspectionResult, Severity
from ..bids_models import DatasetDescription
from ..bids_models._model_globals import _BIDS_RRID


def get_bids_dataset_description(dandiset) -> tuple[DatasetDescription | None, list[InspectionResult]]:
    valid_or_raw: typing.Literal["raw", "valid"] = "valid"
    try:
        metadata = dandiset.get_metadata()
    except pydantic.ValidationError:
        # TODO
        # raw_dandiset_metadata = dandiset.get_raw_metadata()
        valid_or_raw = "raw"

    messages = []
    if valid_or_raw == "valid":
        dataset_description, internal_messages = _get_dataset_description_from_valid_dandiset_metadata(
            metadata=metadata
        )
        messages.extend(internal_messages)
    else:
        dataset_description = None
        message = InspectionResult(
            title="NotImplemented: invalid Dandiset metadata",
            reason="This Dandiset has invalid metadata.",
            solution="`nwb2bids` plans to add support for reading from the raw metadata.",
            category=Category.INTERNAL_ERROR,
            severity=Severity.ERROR,
        )
        messages.append(message)

    return dataset_description, messages


def _get_dataset_description_from_valid_dandiset_metadata(
    metadata: typing.Any,
) -> tuple[DatasetDescription | None, list[InspectionResult]]:
    dataset_description_kwargs = dict()

    dataset_description_kwargs["Name"] = metadata.name
    dataset_description_kwargs["BIDSVersion"] = "1.10"

    messages = []
    if any(data_standard.identifier == _BIDS_RRID for data_standard in metadata.assetsSummary.dataStandard):
        reason = (
            "Dandiset is already organized to the BIDS standard. If only a partial conversion is desired, "
            "please raise an issue on https://github.com/con/nwb2bids/issues/new to discuss the use case."
        )
        message = InspectionResult(
            title="Dandiset is already BIDS",
            reason=reason,
            solution="Skip the conversion of this Dandiset.",
            category=Category.INTERNAL_ERROR,
            severity=Severity.ERROR,
        )
        messages.append(message)
        return None, messages

    if metadata.description is not None:
        dataset_description_kwargs["Description"] = metadata.description
    if metadata.contributor is not None:
        dataset_description_kwargs["Authors"] = [
            contributor.name
            for contributor in metadata.contributor
            for role in contributor.roleName
            if role.value == "dcite:Author"
        ]
    if metadata.license is not None:
        dataset_description_kwargs["License"] = metadata.license[0].value.split(":")[1]

    dataset_description = DatasetDescription(**dataset_description_kwargs)
    return dataset_description, messages



"""Tests for message handling passed to the top-level `converter_nwb_dataset` function."""

import pathlib

import nwb2bids


def test_messages_baseline(minimal_nwbfile_path: pathlib.Path, temporary_bids_directory: pathlib.Path) -> None:
    nwb_paths = [minimal_nwbfile_path]
    messages = nwb2bids.convert_nwb_dataset(nwb_paths=nwb_paths, bids_directory=temporary_bids_directory)

    assert len(messages) == 0


def test_messages_1(problematic_nwbfile_path_1: pathlib.Path, temporary_bids_directory: pathlib.Path) -> None:
    nwb_paths = [problematic_nwbfile_path_1]
    messages = nwb2bids.convert_nwb_dataset(nwb_paths=nwb_paths, bids_directory=temporary_bids_directory)

    expected_messages = [
        nwb2bids.InspectionResult(
            title="Invalid species",
            reason="Participant species is not a proper Latin binomial or NCBI Taxonomy id.",
            solution=(
                "Rename the subject species to a Latin binomial, obolibrary taxonomy link, or NCBI taxonomy "
                "reference."
            ),
            examples=[],
            field="nwbfile.subject.species",
            source_file_paths=nwb_paths,
            data_standards=[nwb2bids.DataStandard.DANDI],
            category=nwb2bids.Category.SCHEMA_INVALIDATION,
            severity=nwb2bids.Severity.ERROR,
        ),
        nwb2bids.InspectionResult(
            title="Invalid participant ID",
            reason=(
                "The participant ID contains invalid characters. BIDS allows only dashes to be used as separators in "
                "subject entity label. Underscores, spaces, slashes, and special characters (including #) are "
                "expressly forbidden."
            ),
            solution="Rename the subject without using spaces or underscores.",
            examples=["`ab_01` -> `ab-01`", "`subject #2` -> `subject-2`", "`id 2 from 9/1/25` -> `id-2-9-1-25`"],
            field="nwbfile.subject.subject_id",
            source_file_paths=nwb_paths,
            data_standards=[nwb2bids.DataStandard.BIDS, nwb2bids.DataStandard.DANDI],
            category=nwb2bids.Category.STYLE_SUGGESTION,
            severity=nwb2bids.Severity.ERROR,
        ),
        nwb2bids.InspectionResult(
            title="Invalid participant sex (BIDS)",
            reason="Participant sex is not one of the allowed patterns by BIDS.",
            solution="Rename the subject sex to be one of the accepted values.",
            examples=["`male` -> `M`", "`Female` -> `F`", "`n/a` -> `U`", "`hermaphrodite` -> `O`"],
            field="nwbfile.subject.sex",
            source_file_paths=nwb_paths,
            data_standards=[nwb2bids.DataStandard.BIDS],
            category=nwb2bids.Category.STYLE_SUGGESTION,
            severity=nwb2bids.Severity.ERROR,
        ),
        nwb2bids.InspectionResult(
            title="Invalid participant sex (archives)",
            reason="Participant sex is not one of the allowed patterns by the common archives.",
            solution="Rename the subject sex to be one of the accepted values.",
            examples=["`male` -> `M`", "`Female` -> `F`", "`n/a` -> `U`", "`hermaphrodite` -> `O`"],
            field="nwbfile.subject.sex",
            source_file_paths=nwb_paths,
            data_standards=[nwb2bids.DataStandard.DANDI],
            category=nwb2bids.Category.STYLE_SUGGESTION,
            severity=nwb2bids.Severity.ERROR,
        ),
    ]
    assert messages == expected_messages


def test_messages_2(problematic_nwbfile_path_2: pathlib.Path, temporary_bids_directory: pathlib.Path) -> None:
    nwb_paths = [problematic_nwbfile_path_2]
    messages = nwb2bids.convert_nwb_dataset(nwb_paths=nwb_paths, bids_directory=temporary_bids_directory)

    expected_messages = [
        nwb2bids.InspectionResult(
            title="Missing participant sex",
            reason="Archives such as DANDI or EMBER require the subject sex to be specified.",
            solution=(
                "Specify the `sex` field of the Subject object attached to the NWB file as one of four options: "
                '"M" (for male), "F" (for female), "U" (for unknown), or "O" (for other).\nNOTE: for certain animal '
                'species with more specific genetic determinants, such as C elegans, use "O" (for other) then further '
                'specify the subtypes using other custom fields. For example, `c_elegans_sex="XO"`'
            ),
            examples=None,
            field="nwbfile.subject.sex",
            source_file_paths=nwb_paths,
            target_file_paths=None,
            data_standards=[nwb2bids.DataStandard.DANDI],
            category=nwb2bids.Category.SCHEMA_INVALIDATION,
            severity=nwb2bids.Severity.CRITICAL,
        ),
        nwb2bids.InspectionResult(
            title="Missing participant species",
            reason="Archives such as DANDI or EMBER require the subject species to be specified.",
            solution=(
                "Specify the `species` field of the Subject object attached to the NWB file as a Latin binomial, "
                "obolibrary taxonomy link, or NCBI taxonomy reference."
            ),
            examples=None,
            field="nwbfile.subject.species",
            source_file_paths=nwb_paths,
            target_file_paths=None,
            data_standards=[nwb2bids.DataStandard.DANDI],
            category=nwb2bids.Category.SCHEMA_INVALIDATION,
            severity=nwb2bids.Severity.CRITICAL,
        ),
        nwb2bids.InspectionResult(
            title="Invalid participant ID",
            reason=(
                "The participant ID contains invalid characters. BIDS allows only dashes to be used as separators in "
                "subject entity label. Underscores, spaces, slashes, and special characters (including #) are "
                "expressly forbidden."
            ),
            solution="Rename the subject without using spaces or underscores.",
            examples=["`ab_01` -> `ab-01`", "`subject #2` -> `subject-2`", "`id 2 from 9/1/25` -> `id-2-9-1-25`"],
            field="nwbfile.subject.subject_id",
            source_file_paths=nwb_paths,
            target_file_paths=None,
            data_standards=[nwb2bids.DataStandard.BIDS, nwb2bids.DataStandard.DANDI],
            category=nwb2bids.Category.STYLE_SUGGESTION,
            severity=nwb2bids.Severity.ERROR,
        ),
    ]
    assert messages == expected_messages


def test_messages_3(problematic_nwbfile_path_3: pathlib.Path, temporary_bids_directory: pathlib.Path) -> None:
    nwb_paths = [problematic_nwbfile_path_3]
    messages = nwb2bids.convert_nwb_dataset(nwb_paths=nwb_paths, bids_directory=temporary_bids_directory)

    expected_messages = [
        nwb2bids.InspectionResult(
            title="Missing participant",
            reason="BIDS requires a subject to be specified for each NWB file.",
            solution="Add a Subject object to each NWB file.",
            examples=None,
            field="nwbfile.subject",
            source_file_paths=nwb_paths,
            target_file_paths=None,
            data_standards=[nwb2bids.DataStandard.BIDS, nwb2bids.DataStandard.DANDI],
            category=nwb2bids.Category.STYLE_SUGGESTION,
            severity=nwb2bids.Severity.CRITICAL,
        )
    ]
    assert messages == expected_messages



import pathlib
import typing

import click

from .._core._convert_nwb_dataset import convert_nwb_dataset


# nwb2bids
@click.group(name="nwb2bids")
def _nwb2bids_cli():
    pass


# nwb2bids convert [OPTIONS] [NWB_PATHS]...
@_nwb2bids_cli.command(name="convert")
@click.argument("nwb_paths", type=str, nargs=-1)
@click.option(
    "--bids-directory",
    "-o",
    help="Path to the folder where the BIDS dataset will be created (default: current working directory).",
    required=False,
    type=click.Path(writable=True),
    default=None,
)
@click.option(
    "--file-mode",
    help=(
        "How to handle the source NWB files when converting to BIDS format. "
        "By default, will attempt to utilize 'symlink' if the system allows. "
        "Otherwise, will 'copy' to preserve directory integrity. "
        "Use 'move' for fastest speeds if you do not wish to keep the original NWB directory structure."
    ),
    required=False,
    type=click.Choice(["copy", "move", "symlink", "auto"], case_sensitive=False),
    default="auto",
)
@click.option(
    "--additional-metadata-file-path",
    "additional_metadata_file_path",
    help=(
        "Path to a JSON file containing additional metadata to be included in the BIDS dataset. "
        "This file should contain a dictionary with keys corresponding to BIDS entities."
    ),
    required=False,
    type=click.Path(exists=True, dir_okay=False, readable=True),
    default=None,
)
@click.option("--silent", "-s", is_flag=True, help="Suppress all console output.", default=False)
def _run_convert_nwb_dataset(
    nwb_paths: tuple[str, ...],
    bids_directory: str | None = None,
    file_mode: typing.Literal["copy", "move", "symlink", "auto"] = "auto",
    additional_metadata_file_path: str | None = None,
    silent: bool = False,
) -> None:
    """
    Convert NWB files to BIDS format.

    NWB_PATHS : A sequence of paths, each pointing to either an NWB file or a directory containing NWB files.
    """
    if len(nwb_paths) == 0:
        message = "Please provide at least one NWB file or directory to convert."
        raise ValueError(message)
    handled_nwb_paths = [pathlib.Path(nwb_path) for nwb_path in nwb_paths]

    messages = convert_nwb_dataset(
        nwb_paths=handled_nwb_paths,
        bids_directory=bids_directory,
        file_mode=file_mode,
        additional_metadata_file_path=additional_metadata_file_path,
    )

    if messages is not None and not silent:
        text = (
            f"{len(messages)} suggestion for improvement was found during conversion."
            if len(messages) == 1
            else f"{len(messages)} suggestions for improvement were found during conversion."
        )
        console_notification = click.style(text=text, fg="yellow")
        click.echo(message=console_notification)



import abc

import pydantic

from .._inspection._inspection_result import InspectionResult


class MutableModel(pydantic.BaseModel):
    """
    Base Pydantic model for all mutable models in `nwb2bids`.

    Parameters
    ----------
    _file_paths : list of str
        Private attribute to store file paths associated with this model.
        Used by the message feature to associate particular issues with particular files.
    """

    model_config = pydantic.ConfigDict(
        validate_assignment=True,  # Re-validate model on mutation
        extra="allow",  # Allow additional custom fields
    )


class BaseMetadataModel(MutableModel):
    """
    Base Pydantic model for all metadata handled by `nwb2bids`.
    """

    messages: list[InspectionResult] = pydantic.Field(
        description="List of auto-detected suggestions.",
        default_factory=list,
        exclude=True,
    )


class BaseMetadataContainerModel(MutableModel, abc.ABC):
    """
    Base Pydantic model for 'containing' some number of other metadata models plus any extra metadata.
    """

    @abc.abstractmethod
    @pydantic.computed_field
    @property
    def messages(self) -> list[InspectionResult]:
        """
        All messages from contained session converters.

        These can accumulate over time based on which instance methods have been called.
        """
        message = "This field should be defined in child classes."
        raise NotImplementedError(message=message)



"""Integration tests for the primary `nwb2bids convert` CLI."""

import pathlib
import subprocess


def test_problematic_cli_messages(problematic_nwbfile_path_1: pathlib.Path, temporary_bids_directory: pathlib.Path):
    command = f"nwb2bids convert {problematic_nwbfile_path_1} -o {temporary_bids_directory}"

    result = subprocess.run(args=command, shell=True, capture_output=True)
    assert result.returncode == 0
    assert b"4 suggestions for improvement were found during conversion." in result.stdout
    assert result.stderr == b""


def test_problematic_cli_silent(problematic_nwbfile_path_1: pathlib.Path, temporary_bids_directory: pathlib.Path):
    command = f"nwb2bids convert {problematic_nwbfile_path_1} -o {temporary_bids_directory} --silent"

    result = subprocess.run(args=command, shell=True, capture_output=True)
    assert result.returncode == 0
    assert result.stdout == b""
    assert result.stderr == b""



import collections
import json
import pathlib
import traceback
import typing

import pandas
import pydantic
import typing_extensions

from ._dandi_utils import get_bids_dataset_description
from ._session_converter import SessionConverter
from .._converters._base_converter import BaseConverter
from .._inspection._inspection_result import Category, InspectionResult, Severity
from ..bids_models import BidsSessionMetadata, DatasetDescription


class DatasetConverter(BaseConverter):
    session_converters: list[SessionConverter] = pydantic.Field(
        description="List of session converters. Typically instantiated by calling `.from_nwb_paths()`."
    )
    dataset_description: DatasetDescription | None = pydantic.Field(
        description="The BIDS-compatible dataset description.",
        default=None,
    )

    @pydantic.computed_field
    @property
    def messages(self) -> list[InspectionResult]:
        """
        All messages from contained session converters.

        These can accumulate over time based on which instance methods have been called.
        """
        messages = [
            message for session_converter in self.session_converters for message in session_converter.messages
        ] + self._internal_messages
        messages.sort(key=lambda message: (-message.category.value, -message.severity.value, message.title))
        return messages

    @classmethod
    @pydantic.validate_call
    def from_remote_dandiset(
        cls,
        dandiset_id: str = pydantic.Field(pattern=r"^\d{6}$"),
        api_url: str | None = None,
        token: str | None = None,
        limit: int | None = None,
    ) -> typing_extensions.Self | None:
        """
        Initialize a converter of a Dandiset to BIDS format.

        Parameters
        ----------
        dandiset_id : str
            The dandiset ID of the Dandiset to be converted.
        api_url : str, optional
            The API URL of a custom DANDI instance to use. If not provided, the API URL of the
            DANDI instance specified by the :envvar:`DANDI_INSTANCE` environment variable
            is used. If the :envvar:`DANDI_INSTANCE` environment variable is not specified,
            The API URL of the `"dandi"` DANDI instance is used.
        token : str, optional
            The authentication token for accessing the DANDI instance.
            If not provided, will attempt to read from the environment variable `DANDI_API_KEY` if it exists.
            This is required for accessing embargoed Dandisets.
        limit : int, optional
            If specified, limits the number of sessions to convert.
            This is mainly useful for testing purposes.
        """
        try:
            import dandi.dandiapi

            client = dandi.dandiapi.DandiAPIClient(api_url=api_url, token=token)
            version_id = "draft"  # Only allow running on draft version
            dandiset = client.get_dandiset(dandiset_id=dandiset_id, version_id=version_id)

            dataset_description, _internal_messages = get_bids_dataset_description(dandiset=dandiset)

            if limit is None:
                assets = list(dandiset.get_assets())
            else:
                assets = [asset for counter, asset in enumerate(dandiset.get_assets()) if counter < limit]

            session_id_to_assets = collections.defaultdict(list)
            for asset in assets:
                asset_metadata = asset.get_raw_metadata()

                for session in asset_metadata.get("wasGeneratedBy", []):
                    if session.get("schemaKey", "") != "Session":
                        continue

                    session_id = session.get("identifier", "")
                    if session_id == "":
                        continue

                    session_id_to_assets[session_id].append(asset)
            sorted_session_id_to_assets = dict(sorted(session_id_to_assets.items(), key=lambda item: item[0]))

            session_converters = [
                SessionConverter(
                    session_id=session_id,
                    nwbfile_paths=[asset.get_content_url(follow_redirects=1, strip_query=True) for asset in assets],
                )
                for session_id, assets in sorted_session_id_to_assets.items()
            ]

            dataset_converter = cls(session_converters=session_converters, dataset_description=dataset_description)
            dataset_converter._internal_messages = _internal_messages
            return dataset_converter
        except Exception:  # noqa
            _internal_messages = [
                InspectionResult(
                    title="Failed to initialize converter on remote Dandiset",
                    reason=(
                        "An error occurred while executing `DatasetConverter.from_remote_dandiset`."
                        f"\n\n{traceback.format_exc()}"
                    ),
                    solution="Please raise an issue on `nwb2bids`: https://github.com/con/nwb2bids/issues.",
                    category=Category.INTERNAL_ERROR,
                    severity=Severity.ERROR,
                )
            ]

            dataset_converter = cls(session_converters=[], dataset_description=None)

        dataset_converter._internal_messages = _internal_messages
        return dataset_converter

    @classmethod
    @pydantic.validate_call
    def from_nwb_paths(
        cls,
        nwb_paths: list[pydantic.FilePath | pydantic.DirectoryPath] = pydantic.Field(min_length=1),
        additional_metadata_file_path: pydantic.FilePath | None = None,
    ) -> typing_extensions.Self:
        """
        Initialize a converter of NWB files to BIDS format.

        Parameters
        ----------
        nwb_paths : iterable of file and directory paths
            An iterable of NWB file paths and directories containing NWB files.
        additional_metadata_file_path : file path, optional
            The path to a JSON file containing additional metadata to be included in the BIDS dataset.

        Returns
        -------
        An instance of DatasetConverter.
        """
        try:
            session_converters = SessionConverter.from_nwb_paths(nwb_paths=nwb_paths)

            dataset_description = None
            if additional_metadata_file_path is not None:
                dataset_description = DatasetDescription.from_file_path(file_path=additional_metadata_file_path)

            session_messages = [
                message
                for session_converter in session_converters
                for message in session_converter.messages
                if session_converter.messages is not None
            ]

            dataset_converter = cls(session_converters=session_converters, dataset_description=dataset_description)
            dataset_converter._internal_messages = session_messages
            return dataset_converter
        except Exception:  # noqa
            _internal_messages = [
                InspectionResult(
                    title="Failed to initialize converter on local NWB files",
                    reason=(
                        "An error occurred while executing `DatasetConverter.from_nwb_paths`."
                        f"\n\n{traceback.format_exc()}"
                    ),
                    solution="Please raise an issue on `nwb2bids`: https://github.com/con/nwb2bids/issues.",
                    category=Category.INTERNAL_ERROR,
                    severity=Severity.ERROR,
                )
            ]

        dataset_converter._internal_messages = _internal_messages
        return dataset_converter

    def extract_metadata(self) -> None:
        try:
            collections.deque(
                (
                    session_converter.extract_metadata()
                    for session_converter in self.session_converters
                    if session_converter.session_metadata is None
                ),
                maxlen=0,
            )
        except Exception:  # noqa
            message = InspectionResult(
                title="Failed to extract metadata for one or more sessions",
                reason=(
                    "An error occurred while executing `DatasetConverter.extract_metadata`."
                    f"\n\n{traceback.format_exc()}"
                ),
                solution="Please raise an issue on `nwb2bids`: https://github.com/con/nwb2bids/issues.",
                category=Category.INTERNAL_ERROR,
                severity=Severity.ERROR,
            )
            self._internal_messages.append(message)

    @pydantic.validate_call
    def convert_to_bids_dataset(
        self,
        bids_directory: str | pathlib.Path | None = None,
        file_mode: typing.Literal["move", "copy", "symlink", "auto"] = "auto",
    ) -> None:
        """
        Convert the directory of NWB files to a BIDS dataset.

        Parameters
        ----------
        bids_directory : directory path, optional
            The path to the directory where the BIDS dataset will be created.
            If not specified, the current working directory will be used if it is valid to become a BIDS dataset.
            To be a valid BIDS dataset, the directory must either start off empty or otherwise contain BIDS-compatible
            files, such as `dataset_description.json`, `participants.tsv`, `README.md`, and so on.
        file_mode : one of "move", "copy", "symlink", or "auto", default: "auto"
            Specifies how to handle the NWB files when converting to BIDS format.
            - "move": Move the files to the BIDS directory.
            - "copy": Copy the files to the BIDS directory.
            - "symlink": Create symbolic links to the files in the BIDS directory.
            - "auto": Decides between all the above based on the system, with preference for linking when possible.
        """
        try:
            bids_directory = self._handle_bids_directory(bids_directory=bids_directory)

            if self.dataset_description is not None:
                self.write_dataset_description(bids_directory=bids_directory)

            self.write_participants_metadata(bids_directory=bids_directory)
            self.write_sessions_metadata(bids_directory=bids_directory)

            generator = (
                session_converter.convert_to_bids_session(bids_directory=bids_directory, file_mode=file_mode)
                for session_converter in self.session_converters
            )
            collections.deque(generator, maxlen=0)
        except Exception:  # noqa
            message = InspectionResult(
                title="Failed to convert to BIDS dataset",
                reason=(
                    "An error occurred while executing `DatasetConverter.convert_to_bids_dataset`."
                    f"\n\n{traceback.format_exc()}"
                ),
                solution="Please raise an issue on `nwb2bids`: https://github.com/con/nwb2bids/issues.",
                category=Category.INTERNAL_ERROR,
                severity=Severity.ERROR,
            )
            self._internal_messages.append(message)

    @pydantic.validate_call
    def write_dataset_description(self, bids_directory: str | pathlib.Path | None = None) -> None:
        bids_directory = self._handle_bids_directory(bids_directory=bids_directory)

        dataset_description_dictionary = self.dataset_description.model_dump()

        dataset_description_file_path = bids_directory / "dataset_description.json"
        with dataset_description_file_path.open(mode="w") as file_stream:
            json.dump(obj=dataset_description_dictionary, fp=file_stream, indent=4)

    @pydantic.validate_call
    def write_participants_metadata(self, bids_directory: str | pathlib.Path | None = None) -> None:
        """
        Write the `participants.tsv` and `participants.json` files.

        Parameters
        ----------
        bids_directory : directory path
            The path to the directory where the BIDS dataset will be created.
        """
        bids_directory = self._handle_bids_directory(bids_directory=bids_directory)

        model_dump_per_session = []
        for session_converter in self.session_converters:
            model_dump = session_converter.session_metadata.participant.model_dump()
            model_dump_per_session.append(model_dump)

        full_participants_data_frame = pandas.DataFrame.from_records(
            data=[
                {key: value for key, value in model_dump.items() if value is not None}
                for model_dump in model_dump_per_session
            ]
        ).astype("string")

        if full_participants_data_frame.empty:
            return

        # Deduplicate all rows of the frame
        deduplicated_participants_data_frame = full_participants_data_frame.drop_duplicates(ignore_index=True)

        # BIDS requires sub- prefix in table values
        participants_data_frame = deduplicated_participants_data_frame.copy()
        participants_data_frame["participant_id"] = participants_data_frame["participant_id"].apply(
            lambda participant_id: f"sub-{participant_id}"
        )
        is_field_in_table = {field: True for field in participants_data_frame.keys()}

        # BIDS Validator is strict regarding column order
        required_column_order = [
            field
            for field in ["participant_id", "species", "sex", "strain"]
            if is_field_in_table.get(field, False) is True
        ]
        column_order = required_column_order + [
            field
            for field in participants_data_frame.columns
            if is_field_in_table.get(field, False) is True and field not in required_column_order
        ]

        participants_tsv_file_path = bids_directory / "participants.tsv"
        participants_data_frame.to_csv(
            path_or_buf=participants_tsv_file_path, mode="w", index=False, sep="\t", columns=column_order
        )

        if len(self.session_converters) > 0:
            is_field_in_table = {field: True for field in participants_data_frame.keys()}
            participants_schema = self.session_converters[0].session_metadata.participant.model_json_schema()
            participants_json = {
                field: info["description"]
                for field, info in participants_schema["properties"].items()
                if is_field_in_table.get(field, False) is True
            }
            participants_json_file_path = bids_directory / "participants.json"
            with participants_json_file_path.open(mode="w") as file_stream:
                json.dump(obj=participants_json, fp=file_stream, indent=4)

    @pydantic.validate_call
    def write_sessions_metadata(self, bids_directory: str | pathlib.Path | None = None) -> None:
        """
        Write the `_sessions.tsv` and `_sessions.json` files, then create empty participant and session directories.

        Parameters
        ----------
        bids_directory : directory path
            The path to the directory where the BIDS dataset will be created.
        """
        bids_directory = self._handle_bids_directory(bids_directory=bids_directory)

        subject_id_to_sessions = collections.defaultdict(list)
        for session_converter in self.session_converters:
            subject_id_to_sessions[session_converter.session_metadata.participant.participant_id].append(
                session_converter
            )

        # TODO: expand beyond just session_id field (mainly via additional metadata)
        sessions_schema = BidsSessionMetadata.model_json_schema()
        sessions_json = {"session_id": sessions_schema["properties"]["session_id"]["description"]}

        for subject_id, sessions_metadata in subject_id_to_sessions.items():
            subject_directory = bids_directory / f"sub-{subject_id}"
            subject_directory.mkdir(exist_ok=True)

            # BIDS requires ses- prefix in table values
            sessions_data_frame = pandas.DataFrame(
                {"session_id": [f"ses-{session_metadata.session_id}" for session_metadata in sessions_metadata]}
            )

            session_tsv_file_path = subject_directory / f"sub-{subject_id}_sessions.tsv"
            sessions_data_frame.to_csv(path_or_buf=session_tsv_file_path, mode="w", index=False, sep="\t")

            session_json_file_path = subject_directory / f"sub-{subject_id}_sessions.json"
            with session_json_file_path.open(mode="w") as file_stream:
                json.dump(obj=sessions_json, fp=file_stream, indent=4)

            for session_metadata in sessions_metadata:
                session_directory = subject_directory / f"ses-{session_metadata.session_id}"
                session_directory.mkdir(exist_ok=True)



import functools
import pathlib
import warnings

import pynwb

warnings.filterwarnings(action="ignore", message="No cached namespaces found in .+")
warnings.filterwarnings(action="ignore", message="Ignoring cached namespace .+")
warnings.filterwarnings(action="ignore", message="Stimulus description 'NA' for .+")
warnings.filterwarnings(action="ignore", message="Use of icephys_filtering is deprecated and will be removed .+")


@functools.cache
def cache_read_nwb(file_path: pathlib.Path) -> pynwb.NWBFile:
    """Cache the read operation per NWB file path to speed up repeated calls."""

    nwbfile = pynwb.read_nwb(path=file_path)
    return nwbfile



import json
import pathlib

import pandas
import pydantic
import pynwb
import typing_extensions

from ..bids_models._base_metadata_model import BaseMetadataModel


class Events(BaseMetadataModel):
    onset: list[float] = pydantic.Field(
        description=(
            "Onset (in seconds) of the event, measured from the beginning of the acquisition of the first data point "
            "stored in the corresponding task data file. Negative onsets are allowed, to account for events that occur "
            "prior to the first stored data point. For example, in case there is a training phase that "
            "begins before the scanning sequence has started, then events from this sequence should have negative "
            "onset time counting down to the beginning of the acquisition of the first recording. If any data "
            "points have been discarded before forming the data file, a time of 0 corresponds to the first stored "
            "data point and not the first acquired data point."
        )
    )
    duration: list[float] = pydantic.Field(
        description=(
            "Duration of the event (measured from onset) in seconds. Must always be either zero or positive "
            '(or n/a if unavailable). A "duration" value of zero implies that the delta function or event is so '
            "short as to be effectively modeled as an impulse."
        ),
    )
    _bids_events_data_frame: pandas.DataFrame
    _nwbfiles: list[pynwb.NWBFile]

    @classmethod
    @pydantic.validate_call
    def from_nwbfiles(cls, nwbfiles: list[pydantic.InstanceOf[pynwb.NWBFile]]) -> typing_extensions.Self | None:
        """
        Extracts all time interval events from the in-memory NWBFile objects.

        Future improvements will include support for non-interval events (ndx-events)
        and DynamicTables with *_time columns.
        """
        if len(nwbfiles) > 1:
            message = "Conversion of multiple NWB files per session is not yet supported."
            raise NotImplementedError(message)
        nwbfile = nwbfiles[0]

        nwb_events_data_frame = _get_events_data_frame(nwbfile=nwbfile)
        if nwb_events_data_frame is None:
            return None

        # Collapse 'start_time' and 'stop_time' columns into 'onset' and 'duration' columns
        bids_events_data_frame = nwb_events_data_frame.copy()
        bids_events_data_frame["duration"] = bids_events_data_frame["stop_time"] - bids_events_data_frame["start_time"]
        bids_events_data_frame = bids_events_data_frame.rename(columns={"start_time": "onset"})
        bids_events_data_frame = bids_events_data_frame.drop(columns=["stop_time"])
        bids_events_data_frame = bids_events_data_frame.sort_values(
            by=["onset", "duration"], ascending=[True, False]
        ).reset_index(drop=True)

        dictionary = {
            "onset": bids_events_data_frame["onset"].tolist(),
            "duration": bids_events_data_frame["duration"].tolist(),
        }
        custom_fields = set(bids_events_data_frame.columns) - {"onset", "duration"}
        dictionary.update({custom_field: list(bids_events_data_frame[custom_field]) for custom_field in custom_fields})
        dictionary["_bids_events_data_frame"] = bids_events_data_frame
        dictionary["_nwbfiles"] = nwbfiles

        bids_events = cls(**dictionary)
        return bids_events

    @pydantic.validate_call
    def to_tsv(self, file_path: str | pathlib.Path) -> None:
        """
        Write the BIDS events table to a TSV file.

        Parameters
        ----------
        file_path : str or pathlib.Path
            The path to the output TSV file.
        """
        bids_events_data_frame = self.model_extra["_bids_events_data_frame"]

        required_column_order = ["onset", "duration", "nwb_table"]
        column_order = required_column_order + [
            column for column in bids_events_data_frame.columns if column not in required_column_order
        ]

        bids_events_data_frame.to_csv(path_or_buf=file_path, sep="\t", index=False, columns=column_order)

    @pydantic.validate_call
    def to_json(self, file_path: str | pathlib.Path) -> None:
        """
        Write the BIDS events JSON sidecar file.

        Parameters
        ----------
        file_path : str or pathlib.Path
            The path to the output JSON file.
        """
        nwbfile = self.model_extra["_nwbfiles"][0]

        fields_metadata = _get_events_metadata(nwbfile=nwbfile)

        with file_path.open(mode="w") as file_stream:
            json.dump(obj=fields_metadata, fp=file_stream, indent=4)


def _get_events_data_frame(nwbfile: pynwb.NWBFile) -> pandas.DataFrame | None:
    """
    Extracts all time interval events from the NWB file and returns them as a single data frame.

    Future improvements will include support for non-interval events (ndx-events) and DynamicTables with *_time columns.
    """
    time_intervals = _get_all_time_intervals(nwbfile=nwbfile)
    if time_intervals is None:
        return None

    time_interval_names = [time_interval.name for time_interval in time_intervals]
    if len(set(time_interval_names)) != len(time_interval_names):
        message = (
            f"\nFound duplicate time interval names in the NWB file: {time_interval_names}\n"
            "Please raise an issue at https://github.com/con/nwb2bids/issues/new.\n\n"
        )
        raise ValueError(message)

    all_column_names = {column_name: True for time_interval in time_intervals for column_name in time_interval.colnames}
    if all_column_names.get("nwb_table", None) is not None:
        message = (
            "\nA column with the name 'nwb_table' was found in the NWB file.\n"
            "This is reserved for the nwb2bids conversion process and will require an override to proceed.\n"
            "Please raise an issue at https://github.com/con/nwb2bids/issues/new.\n\n"
        )
        raise ValueError(message)

    all_data_frames = [time_interval.to_dataframe() for time_interval in time_intervals]
    for index, time_interval in enumerate(time_intervals):
        all_data_frames[index]["nwb_table"] = time_interval.name
        if "timeseries" in all_data_frames[index].columns:
            all_data_frames[index] = all_data_frames[index].drop(columns=["timeseries"])

    events_table = pandas.concat(objs=all_data_frames, ignore_index=True)
    return events_table


def _get_all_time_intervals(
    nwbfile: pynwb.NWBFile,
) -> list[pynwb.epoch.TimeIntervals] | None:
    """
    Extracts all time interval events from the NWB file and returns them as a list of TimeIntervals objects.
    """
    time_intervals: list[pynwb.epoch.TimeIntervals] = [
        neurodata_object
        for neurodata_object in nwbfile.acquisition.values()
        if isinstance(neurodata_object, pynwb.epoch.TimeIntervals)
    ]
    if nwbfile.trials is not None:
        time_intervals.append(nwbfile.trials)
    if nwbfile.epochs is not None:
        time_intervals.append(nwbfile.epochs)

    if len(time_intervals) == 0:
        return None

    return time_intervals


def _get_events_metadata(nwbfile: pynwb.NWBFile) -> dict | None:
    time_intervals = _get_all_time_intervals(nwbfile=nwbfile)
    if time_intervals is None:
        return None

    time_interval_names = [time_interval.name for time_interval in time_intervals]

    common_nwb_table_hed = {
        "trials": "Experimental-trial",
        "epochs": "Time-block",
    }

    event_metadata = {
        time_interval.name: {"Description": time_interval.description}
        for time_interval in time_intervals
        if time_interval.description
    }

    # Follow-up TODO: assign HED tags based on neurodata type once extended beyond TimeIntervals
    event_metadata["nwb_table"] = {
        "nwb_table": {
            "Description": "The name of the NWB table from which this event was extracted.",
            "Levels": {table_name: f"The '{table_name}' table in the NWB file." for table_name in time_interval_names},
            "HED": {
                table_name: common_nwb_table_hed.get(table_name, "Time-interval") for table_name in time_interval_names
            },
        }
    }

    return event_metadata



"""
Including these directly within the top-level `__init__.py` makes them visible to autocompletion.

But we only want the imports to trigger, not for them to actually be exposed.
"""

from .testing import assert_subdirectory_structure

_hide = True



"""Integration tests for the primary `convert_nwb_dataset` function."""

import pathlib

import nwb2bids


def test_minimal_convert_nwb_dataset_from_directory(
    minimal_nwbfile_path: pathlib.Path, temporary_bids_directory: pathlib.Path
):
    nwb_paths = [minimal_nwbfile_path.parent]
    nwb2bids.convert_nwb_dataset(nwb_paths=nwb_paths, bids_directory=temporary_bids_directory)

    expected_structure = {
        temporary_bids_directory: {
            "directories": {"sub-123"},
            "files": {"dataset_description.json", "participants.json", "participants.tsv"},
        },
        temporary_bids_directory
        / "sub-123": {
            "directories": {"ses-456"},
            "files": {"sub-123_sessions.json", "sub-123_sessions.tsv"},
        },
        temporary_bids_directory
        / "sub-123"
        / "ses-456": {
            "directories": {"ecephys"},
            "files": set(),
        },
        temporary_bids_directory
        / "sub-123"
        / "ses-456"
        / "ecephys": {
            "directories": set(),
            "files": {
                "sub-123_ses-456_ecephys.nwb",
            },
        },
    }
    nwb2bids.testing.assert_subdirectory_structure(
        directory=temporary_bids_directory, expected_structure=expected_structure
    )


def test_minimal_convert_nwb_dataset_from_file_path(
    minimal_nwbfile_path: pathlib.Path, temporary_bids_directory: pathlib.Path
):
    nwb_paths = [minimal_nwbfile_path]
    nwb2bids.convert_nwb_dataset(nwb_paths=nwb_paths, bids_directory=temporary_bids_directory)

    expected_structure = {
        temporary_bids_directory: {
            "directories": {"sub-123"},
            "files": {"dataset_description.json", "participants.json", "participants.tsv"},
        },
        temporary_bids_directory
        / "sub-123": {
            "directories": {"ses-456"},
            "files": {"sub-123_sessions.json", "sub-123_sessions.tsv"},
        },
        temporary_bids_directory
        / "sub-123"
        / "ses-456": {
            "directories": {"ecephys"},
            "files": set(),
        },
        temporary_bids_directory
        / "sub-123"
        / "ses-456"
        / "ecephys": {
            "directories": set(),
            "files": {
                "sub-123_ses-456_ecephys.nwb",
            },
        },
    }
    nwb2bids.testing.assert_subdirectory_structure(
        directory=temporary_bids_directory, expected_structure=expected_structure
    )


def test_ecephys_convert_nwb_dataset(ecephys_nwbfile_path: pathlib.Path, temporary_bids_directory: pathlib.Path):
    nwb_paths = [ecephys_nwbfile_path]
    nwb2bids.convert_nwb_dataset(nwb_paths=nwb_paths, bids_directory=temporary_bids_directory)

    expected_structure = {
        temporary_bids_directory: {
            "directories": {"sub-123"},
            "files": {"dataset_description.json", "participants.json", "participants.tsv"},
        },
        temporary_bids_directory
        / "sub-123": {
            "directories": {"ses-789"},
            "files": {"sub-123_sessions.json", "sub-123_sessions.tsv"},
        },
        temporary_bids_directory
        / "sub-123"
        / "ses-789": {
            "directories": {"ecephys"},
            "files": set(),
        },
        temporary_bids_directory
        / "sub-123"
        / "ses-789"
        / "ecephys": {
            "directories": set(),
            "files": {
                "sub-123_ses-789_ecephys.nwb",
                "sub-123_ses-789_channels.tsv",
                "sub-123_ses-789_channels.json",
                "sub-123_ses-789_electrodes.tsv",
                "sub-123_ses-789_electrodes.json",
                "sub-123_ses-789_probes.tsv",
                "sub-123_ses-789_probes.json",
            },
        },
    }
    nwb2bids.testing.assert_subdirectory_structure(
        directory=temporary_bids_directory, expected_structure=expected_structure
    )


def test_optional_bids_directory(
    minimal_nwbfile_path: pathlib.Path, temporary_bids_directory: pathlib.Path, monkeypatch
):
    new_bids_directory = temporary_bids_directory / "bids"
    monkeypatch.chdir(temporary_bids_directory)

    nwb_paths = [minimal_nwbfile_path]
    nwb2bids.convert_nwb_dataset(nwb_paths=nwb_paths)

    expected_structure = {
        new_bids_directory: {
            "directories": {"sub-123"},
            "files": {"dataset_description.json", "participants.json", "participants.tsv"},
        },
        new_bids_directory
        / "sub-123": {
            "directories": {"ses-456"},
            "files": {"sub-123_sessions.json", "sub-123_sessions.tsv"},
        },
        new_bids_directory
        / "sub-123"
        / "ses-456": {
            "directories": {"ecephys"},
            "files": set(),
        },
        new_bids_directory
        / "sub-123"
        / "ses-456"
        / "ecephys": {
            "directories": set(),
            "files": {
                "sub-123_ses-456_ecephys.nwb",
            },
        },
    }
    nwb2bids.testing.assert_subdirectory_structure(directory=new_bids_directory, expected_structure=expected_structure)



"""Integration tests for the primary `convert_nwb_dataset` function with events data."""

import pathlib

import nwb2bids


def test_trials_events(trials_events_nwbfile_path: pathlib.Path, temporary_bids_directory: pathlib.Path):
    nwb_paths = [trials_events_nwbfile_path]
    nwb2bids.convert_nwb_dataset(nwb_paths=nwb_paths, bids_directory=temporary_bids_directory)

    expected_structure = {
        temporary_bids_directory: {
            "directories": {"sub-123"},
            "files": {"dataset_description.json", "participants.json", "participants.tsv"},
        },
        temporary_bids_directory
        / "sub-123": {
            "directories": {"ses-456"},
            "files": {"sub-123_sessions.json", "sub-123_sessions.tsv"},
        },
        temporary_bids_directory
        / "sub-123"
        / "ses-456": {
            "directories": {"ecephys"},
            "files": set(),
        },
        temporary_bids_directory
        / "sub-123"
        / "ses-456"
        / "ecephys": {
            "directories": set(),
            "files": {
                "sub-123_ses-456_ecephys.nwb",
                "sub-123_ses-456_events.tsv",
                "sub-123_ses-456_events.json",
            },
        },
    }
    nwb2bids.testing.assert_subdirectory_structure(
        directory=temporary_bids_directory, expected_structure=expected_structure
    )


def test_epochs_events(epochs_events_nwbfile_path: pathlib.Path, temporary_bids_directory: pathlib.Path):
    nwb_paths = [epochs_events_nwbfile_path]
    nwb2bids.convert_nwb_dataset(nwb_paths=nwb_paths, bids_directory=temporary_bids_directory)

    expected_structure = {
        temporary_bids_directory: {
            "directories": {"sub-123"},
            "files": {"dataset_description.json", "participants.json", "participants.tsv"},
        },
        temporary_bids_directory
        / "sub-123": {
            "directories": {"ses-456"},
            "files": {"sub-123_sessions.json", "sub-123_sessions.tsv"},
        },
        temporary_bids_directory
        / "sub-123"
        / "ses-456": {
            "directories": {"ecephys"},
            "files": set(),
        },
        temporary_bids_directory
        / "sub-123"
        / "ses-456"
        / "ecephys": {
            "directories": set(),
            "files": {
                "sub-123_ses-456_ecephys.nwb",
                "sub-123_ses-456_events.tsv",
                "sub-123_ses-456_events.json",
            },
        },
    }
    nwb2bids.testing.assert_subdirectory_structure(
        directory=temporary_bids_directory, expected_structure=expected_structure
    )


def test_multiple_events(multiple_events_nwbfile_path: pathlib.Path, temporary_bids_directory: pathlib.Path):
    nwb_paths = [multiple_events_nwbfile_path]
    nwb2bids.convert_nwb_dataset(nwb_paths=nwb_paths, bids_directory=temporary_bids_directory)

    expected_structure = {
        temporary_bids_directory: {
            "directories": {"sub-123"},
            "files": {"dataset_description.json", "participants.json", "participants.tsv"},
        },
        temporary_bids_directory
        / "sub-123": {
            "directories": {"ses-456"},
            "files": {"sub-123_sessions.json", "sub-123_sessions.tsv"},
        },
        temporary_bids_directory
        / "sub-123"
        / "ses-456": {
            "directories": {"ecephys"},
            "files": set(),
        },
        temporary_bids_directory
        / "sub-123"
        / "ses-456"
        / "ecephys": {
            "directories": set(),
            "files": {
                "sub-123_ses-456_ecephys.nwb",
                "sub-123_ses-456_events.tsv",
                "sub-123_ses-456_events.json",
            },
        },
    }
    nwb2bids.testing.assert_subdirectory_structure(
        directory=temporary_bids_directory, expected_structure=expected_structure
    )



import pathlib
import re

import h5py
import pydantic
import pynwb
import typing_extensions

from ._base_metadata_model import BaseMetadataContainerModel
from ._channels import ChannelTable
from ._electrodes import ElectrodeTable
from ._events import Events
from ._model_globals import _VALID_ID_REGEX
from ._participant import Participant
from ._probes import ProbeTable
from .._inspection._inspection_result import Category, DataStandard, InspectionResult, Severity
from .._tools import cache_read_nwb


class BidsSessionMetadata(BaseMetadataContainerModel):
    """
    Schema for the metadata of a single BIDS session.
    """

    session_id: str | None = pydantic.Field(description="A unique session identifier.", default=None)
    participant: Participant = pydantic.Field(description="Metadata about a participant used in this experiment.")
    events: Events | None = pydantic.Field(
        description="Timing data and metadata regarding events that occur during this experiment.", default=None
    )
    probe_table: ProbeTable | None = None
    channel_table: ChannelTable | None = None
    electrode_table: ElectrodeTable | None = None

    @pydantic.computed_field
    @property
    def messages(self) -> list[InspectionResult]:
        messages = self.participant.messages.copy()
        messages += self._internal_messages
        if self.events is not None:
            messages += self.events.messages
        if self.probe_table is not None:
            messages += self.probe_table.messages
        if self.channel_table is not None:
            messages += self.channel_table.messages
        if self.electrode_table is not None:
            messages += self.electrode_table.messages
        messages.sort(key=lambda message: (-message.category.value, -message.severity.value, message.title))
        return messages

    def _check_fields(self, file_paths: list[pathlib.Path] | list[pydantic.HttpUrl]) -> None:
        # Check if values are specified
        internal_messages = []
        if self.session_id is None:
            internal_messages.append(
                InspectionResult(
                    title="Missing session ID",
                    reason="A unique ID is required for all individual sessions.",
                    solution="Specify the `session_id` field of the NWB file object.",
                    field="nwbfile.session_id",
                    source_file_paths=file_paths,
                    data_standards=[DataStandard.BIDS, DataStandard.DANDI],
                    category=Category.SCHEMA_INVALIDATION,
                    severity=Severity.CRITICAL,
                )
            )

        # Check if specified values are valid
        if self.session_id is not None and re.match(pattern=_VALID_ID_REGEX, string=self.session_id) is None:
            internal_messages.append(
                InspectionResult(
                    title="Invalid session ID",
                    reason=(
                        "The session ID contains invalid characters. "
                        "BIDS allows only dashes to be used as separators in session entity label. "
                        "Underscores, spaces, slashes, and special characters (including #) are expressly forbidden."
                    ),
                    solution="Rename the session without using spaces or underscores.",
                    examples=[
                        "`ses_01` -> `ses-01`",
                        "`session #2` -> `session-2`",
                        "`id 2 from 9/1/25` -> `id-2-9-1-25`",
                    ],
                    field="nwbfile.session_id",
                    source_file_paths=file_paths,
                    data_standards=[DataStandard.BIDS, DataStandard.DANDI],
                    category=Category.STYLE_SUGGESTION,
                    severity=Severity.ERROR,
                )
            )
        self._internal_messages = internal_messages

    @classmethod
    @pydantic.validate_call
    def from_nwbfile_paths(
        cls, nwbfile_paths: list[pydantic.FilePath] | list[pydantic.HttpUrl] = pydantic.Field(min_length=1)
    ) -> typing_extensions.Self:
        # Differentiate local path from URL
        if isinstance(next(iter(nwbfile_paths)), pathlib.Path):
            nwbfiles = [cache_read_nwb(nwbfile_path) for nwbfile_path in nwbfile_paths]
        else:
            nwbfiles = [_stream_nwb(url=url) for url in nwbfile_paths]

        session_ids = {nwbfile.session_id for nwbfile in nwbfiles}
        if len(session_ids) > 1:
            message = "Multiple differing session IDs found - please check how this method was called."
            raise ValueError(message)
        session_id = next(iter(session_ids))

        participant = Participant.from_nwbfiles(nwbfiles=nwbfiles)
        events = Events.from_nwbfiles(nwbfiles=nwbfiles)
        probe_table = ProbeTable.from_nwbfiles(nwbfiles=nwbfiles)
        channel_table = ChannelTable.from_nwbfiles(nwbfiles=nwbfiles)
        electrode_table = ElectrodeTable.from_nwbfiles(nwbfiles=nwbfiles)

        dictionary = {
            "session_id": session_id,
            "participant": participant,
        }
        if events is not None:
            dictionary["events"] = events
        if probe_table is not None:
            dictionary["probe_table"] = probe_table
        if channel_table is not None:
            dictionary["channel_table"] = channel_table
        if electrode_table is not None:
            dictionary["electrode_table"] = electrode_table

        session_metadata = cls(**dictionary)
        session_metadata._check_fields(file_paths=nwbfile_paths)
        return session_metadata


def _stream_nwb(url: pydantic.HttpUrl) -> pynwb.NWBFile:
    """
    Stream an NWB file from a URL using remfile.

    Parameters
    ----------
    url : pydantic.HttpUrl
        The URL of the NWB file to stream.

    Returns
    -------
    pynwb.NWBFile
        The streamed NWB file.
    """
    import remfile

    rem_file = remfile.File(url=str(url))

    try:
        h5py_file = h5py.File(name=rem_file, mode="r")
    except Exception as exception:
        message = (
            f"\nFailed to open NWB file from URL {url}: {exception}\n\n"
            "Possible that backend is not supported.\n"
            "Please raise an issue on https://github.com/con/nwb2bids/issues/new to discuss."
        )
        raise ValueError(message)

    file_io = pynwb.NWBHDF5IO(file=h5py_file, mode="r")
    nwbfile = file_io.read()
    return nwbfile



import os
import pathlib
import typing

import pydantic


@pydantic.validate_call
def assert_subdirectory_structure(
    *,
    directory: pathlib.Path,
    expected_structure: dict[pathlib.Path, dict[typing.Literal["directories", "files"], set[str]]],
) -> None:
    """
    Assert that the subdirectory structure matches the expected structure.

    Parameters
    ----------
    directory : directory path
        The path to the directory whose structure is to be checked.
    expected_structure : dict
        A dictionary representing the expected structure of the subdirectory.
        Keys are the full paths per expected subdirectory.
        Values are dictionaries with required 'directories' and 'files' keys whose values are sets of string names for
        all expected entities at that level.
    """
    # Future TODO: adjust to pathlib.Path.walk once 3.12 is minimum
    for subdirectory_path, directories, files in os.walk(top=directory):
        subdirectory_path = pathlib.Path(subdirectory_path)
        expected = expected_structure.get(subdirectory_path, None)

        assert expected is not None, (
            f"\n\nUnexpected subdirectory {subdirectory_path}.\n\n"
            f"Expected subdirectories: {list(expected_structure.keys())}\n"
        )
        assert set(expected.keys()) == {"directories", "files"}, (
            f"\n\nUnexpected keys in expected structure for {subdirectory_path}.\n\n"
            f"Expected keys: 'directories', 'files'\n"
            f"Found keys: {set(expected.keys())}\n\n"
        )
        assert set(directories) == set(expected["directories"]), (
            f"\n\nUnexpected directories in {subdirectory_path}.\n\n"
            f"Expected: {expected['directories']}\n"
            f"Found: {set(directories)}\n"
            f"Difference: {set(directories) - set(expected['directories'])}\n"
            f"Extra: {set(expected['directories']) - set(directories)}\n\n"
        )
        assert set(files) == set(expected["files"]), (
            f"\n\nUnexpected files in {subdirectory_path}.\n\n"
            f"Expected: {expected['files']}\n"
            f"Found: {files}\n"
            f"Difference: {set(files) - set(expected['files'])}\n"
            f"Extra: {set(expected['files']) - set(files)}\n\n"
        )



"""Integration tests for the primary `nwb2bids convert` CLI."""

import pathlib
import subprocess

import nwb2bids


def test_minimal_cli_on_directory(minimal_nwbfile_path: pathlib.Path, temporary_bids_directory: pathlib.Path):
    command = f"nwb2bids convert {minimal_nwbfile_path.parent} -o {temporary_bids_directory}"

    result = subprocess.run(args=command, shell=True, capture_output=True)
    assert (
        result.returncode == 0
    ), f"\n\nCLI command failed with:\nStandard Output: {result.stdout}\nStandard Error: {result.stderr}\n\n"

    expected_structure = {
        temporary_bids_directory: {
            "directories": {"sub-123"},
            "files": {"dataset_description.json", "participants.json", "participants.tsv"},
        },
        temporary_bids_directory
        / "sub-123": {
            "directories": {"ses-456"},
            "files": {"sub-123_sessions.json", "sub-123_sessions.tsv"},
        },
        temporary_bids_directory
        / "sub-123"
        / "ses-456": {
            "directories": {"ecephys"},
            "files": set(),
        },
        temporary_bids_directory
        / "sub-123"
        / "ses-456"
        / "ecephys": {
            "directories": set(),
            "files": {
                "sub-123_ses-456_ecephys.nwb",
            },
        },
    }
    nwb2bids.testing.assert_subdirectory_structure(
        directory=temporary_bids_directory, expected_structure=expected_structure
    )


def test_minimal_cli_on_file_path(minimal_nwbfile_path: pathlib.Path, temporary_bids_directory: pathlib.Path):
    command = f"nwb2bids convert --bids-directory {temporary_bids_directory} {minimal_nwbfile_path}"

    result = subprocess.run(args=command, shell=True, capture_output=True)
    assert (
        result.returncode == 0
    ), f"\n\nCLI command failed with:\nStandard Output: {result.stdout}\nStandard Error: {result.stderr}\n\n"

    expected_structure = {
        temporary_bids_directory: {
            "directories": {"sub-123"},
            "files": {"dataset_description.json", "participants.json", "participants.tsv"},
        },
        temporary_bids_directory
        / "sub-123": {
            "directories": {"ses-456"},
            "files": {"sub-123_sessions.json", "sub-123_sessions.tsv"},
        },
        temporary_bids_directory
        / "sub-123"
        / "ses-456": {
            "directories": {"ecephys"},
            "files": set(),
        },
        temporary_bids_directory
        / "sub-123"
        / "ses-456"
        / "ecephys": {
            "directories": set(),
            "files": {
                "sub-123_ses-456_ecephys.nwb",
            },
        },
    }
    nwb2bids.testing.assert_subdirectory_structure(
        directory=temporary_bids_directory, expected_structure=expected_structure
    )


def test_ecephys_cli(ecephys_nwbfile_path: pathlib.Path, temporary_bids_directory: pathlib.Path):
    command = f"nwb2bids convert {ecephys_nwbfile_path.parent} -o {temporary_bids_directory}"
    result = subprocess.run(args=command, shell=True, capture_output=True)
    assert (
        result.returncode == 0
    ), f"\n\nCLI command failed with:\nStandard Output: {result.stdout}\nStandard Error: {result.stderr}\n\n"

    expected_structure = {
        temporary_bids_directory: {
            "directories": {"sub-123"},
            "files": {"dataset_description.json", "participants.json", "participants.tsv"},
        },
        temporary_bids_directory
        / "sub-123": {
            "directories": {"ses-789"},
            "files": {"sub-123_sessions.json", "sub-123_sessions.tsv"},
        },
        temporary_bids_directory
        / "sub-123"
        / "ses-789": {
            "directories": {"ecephys"},
            "files": set(),
        },
        temporary_bids_directory
        / "sub-123"
        / "ses-789"
        / "ecephys": {
            "directories": set(),
            "files": {
                "sub-123_ses-789_ecephys.nwb",
                "sub-123_ses-789_channels.tsv",
                "sub-123_ses-789_channels.json",
                "sub-123_ses-789_electrodes.tsv",
                "sub-123_ses-789_electrodes.json",
                "sub-123_ses-789_probes.tsv",
                "sub-123_ses-789_probes.json",
            },
        },
    }
    nwb2bids.testing.assert_subdirectory_structure(
        directory=temporary_bids_directory, expected_structure=expected_structure
    )


def test_minimal_cli_on_file_paths(
    directory_with_multiple_nwbfiles: pathlib.Path, temporary_bids_directory: pathlib.Path
):
    command = f"nwb2bids convert -o {temporary_bids_directory} {directory_with_multiple_nwbfiles}/*.nwb"
    result = subprocess.run(args=command, shell=True, capture_output=True)
    assert (
        result.returncode == 0
    ), f"\n\nCLI command failed with:\nStandard Output: {result.stdout}\nStandard Error: {result.stderr}\n\n"

    expected_structure = {
        temporary_bids_directory: {
            "directories": {"sub-123"},
            "files": {"dataset_description.json", "participants.json", "participants.tsv"},
        },
        temporary_bids_directory
        / "sub-123": {
            "directories": {"ses-session-0", "ses-session-1"},
            "files": {"sub-123_sessions.json", "sub-123_sessions.tsv"},
        },
        temporary_bids_directory
        / "sub-123"
        / "ses-session-0": {
            "directories": {"ecephys"},
            "files": set(),
        },
        temporary_bids_directory
        / "sub-123"
        / "ses-session-0"
        / "ecephys": {
            "directories": set(),
            "files": {
                "sub-123_ses-session-0_ecephys.nwb",
            },
        },
        temporary_bids_directory
        / "sub-123"
        / "ses-session-1": {
            "directories": {"ecephys"},
            "files": set(),
        },
        temporary_bids_directory
        / "sub-123"
        / "ses-session-1"
        / "ecephys": {
            "directories": set(),
            "files": {
                "sub-123_ses-session-1_ecephys.nwb",
            },
        },
    }
    nwb2bids.testing.assert_subdirectory_structure(
        directory=temporary_bids_directory, expected_structure=expected_structure
    )



import pathlib
import subprocess


def test_cli_without_arg(temporary_bids_directory: pathlib.Path):
    command = f"nwb2bids convert -o {temporary_bids_directory}"

    result = subprocess.run(args=command, shell=True, capture_output=True)
    assert result.returncode == 1
    assert result.stdout == b""
    assert b"Please provide at least one NWB file or directory to convert." in result.stderr



"""Unit tests for the `DatasetConverter` class and its methods."""

import pathlib

import pydantic
import pytest

import nwb2bids


@pytest.mark.remote
def test_remote_dataset_converter_initialization(temporary_bids_directory: pathlib.Path):
    dataset_converter = nwb2bids.DatasetConverter.from_remote_dandiset(dandiset_id="000003", limit=2)
    assert isinstance(dataset_converter, nwb2bids.DatasetConverter)


@pytest.mark.remote
def test_remote_dataset_converter_metadata_extraction(temporary_bids_directory: pathlib.Path):
    dataset_converter = nwb2bids.DatasetConverter.from_remote_dandiset(dandiset_id="000003", limit=2)
    dataset_converter.extract_metadata()

    assert len(dataset_converter.session_converters) == 2

    session_converter = dataset_converter.session_converters[0]
    assert session_converter.session_id == "YutaMouse20-140321"
    assert session_converter.nwbfile_paths == [
        pydantic.HttpUrl("https://dandiarchive.s3.amazonaws.com/blobs/bb8/1f7/bb81f7b3-4cfa-40e7-aa89-95beb1954d8c")
    ]

    session_metadata = session_converter.session_metadata
    assert session_metadata.participant == nwb2bids.bids_models.Participant(
        messages=[
            nwb2bids.InspectionResult(
                title="Missing participant sex",
                reason="Archives such as DANDI or EMBER require the subject sex to be specified.",
                solution=(
                    "Specify the `sex` field of the Subject object attached to the NWB file as one of four options: "
                    '"M" (for male), "F" (for female), "U" (for unknown), or "O" (for other).\nNOTE: for certain '
                    'animal species with more specific genetic determinants, such as C elegans, use "O" (for other) '
                    'then further specify the subtypes using other custom fields. For example, `c_elegans_sex="XO"`'
                ),
                examples=None,
                field="nwbfile.subject.sex",
                source_file_paths=None,
                target_file_paths=None,
                data_standards=[nwb2bids.DataStandard.DANDI],
                category=nwb2bids.Category.SCHEMA_INVALIDATION,
                severity=nwb2bids.Severity.CRITICAL,
            ),
        ],
        participant_id="YutaMouse20",
        species="Mus musculus",
        sex=None,
        strain=None,
    )
    assert session_metadata.events is None

    assert session_metadata.probe_table == nwb2bids.bids_models.ProbeTable(
        probes=[
            nwb2bids.bids_models.Probe(
                probe_id="Implant", type=None, description="Silicon electrodes on Intan probe.", manufacturer=None
            )
        ]
    )

    assert session_metadata.channel_table is not None
    assert len(session_metadata.channel_table.channels) == 65
    assert session_metadata.channel_table.channels[0] == nwb2bids.bids_models.Channel(
        channel_id="0", electrode_id="0", type="EXT", unit="V", sampling_frequency=None, gain=None
    )

    assert session_metadata.electrode_table is not None
    assert len(session_metadata.channel_table.channels) == 65
    assert session_metadata.electrode_table.electrodes[0] == nwb2bids.bids_models.Electrode(
        electrode_id=0, probe_id="Implant", location="unknown"
    )



"""
NWB2BIDS
========

Reorganize NWB files into a BIDS directory layout.
"""

from ._core._convert_nwb_dataset import convert_nwb_dataset
from ._converters._dataset_converter import DatasetConverter
from ._converters._session_converter import SessionConverter
from ._inspection._inspection_result import InspectionResult, Severity, Category, DataStandard

__all__ = [
    # Public methods and classes
    "convert_nwb_dataset",
    "Category",
    "DatasetConverter",
    "SessionConverter",
    "Severity",
    "DataStandard",
    "InspectionResult",
    # Public submodules
    "bids_models",
    "testing",
]

# Trigger import of hidden submodule elements (only need to import one item to trigger the rest)
from ._hidden_top_level_imports import _hide



"""Unit tests for the `DatasetConverter` class and its methods."""

import json
import pathlib

import pandas

import nwb2bids


def test_dataset_converter_directory_initialization(
    minimal_nwbfile_path: pathlib.Path, temporary_bids_directory: pathlib.Path
):
    nwb_paths = [minimal_nwbfile_path.parent]
    dataset_converter = nwb2bids.DatasetConverter.from_nwb_paths(nwb_paths=nwb_paths)

    assert isinstance(dataset_converter, nwb2bids.DatasetConverter)
    assert isinstance(dataset_converter.session_converters, list)
    assert len(dataset_converter.session_converters) == 1
    assert isinstance(dataset_converter.session_converters[0], nwb2bids.SessionConverter)


def test_dataset_converter_file_paths_initialization(
    minimal_nwbfile_path: pathlib.Path, temporary_bids_directory: pathlib.Path
):
    nwb_paths = [minimal_nwbfile_path]
    dataset_converter = nwb2bids.DatasetConverter.from_nwb_paths(nwb_paths=nwb_paths)

    assert isinstance(dataset_converter, nwb2bids.DatasetConverter)
    assert isinstance(dataset_converter.session_converters, list)
    assert len(dataset_converter.session_converters) == 1
    assert isinstance(dataset_converter.session_converters[0], nwb2bids.SessionConverter)


def test_dataset_converter_both_file_and_directory_initialization(
    minimal_nwbfile_path: pathlib.Path, ecephys_nwbfile_path: pathlib.Path, temporary_bids_directory: pathlib.Path
):
    nwb_paths = [minimal_nwbfile_path.parent, ecephys_nwbfile_path]
    dataset_converter = nwb2bids.DatasetConverter.from_nwb_paths(nwb_paths=nwb_paths)

    assert isinstance(dataset_converter, nwb2bids.DatasetConverter)
    assert isinstance(dataset_converter.session_converters, list)
    assert len(dataset_converter.session_converters) == 2
    assert all(isinstance(converter, nwb2bids.SessionConverter) for converter in dataset_converter.session_converters)


def test_dataset_converter_metadata_extraction(
    minimal_nwbfile_path: pathlib.Path, temporary_bids_directory: pathlib.Path
):
    nwb_paths = [minimal_nwbfile_path]
    dataset_converter = nwb2bids.DatasetConverter.from_nwb_paths(nwb_paths=nwb_paths)
    dataset_converter.extract_metadata()

    expected_session_converters = [
        nwb2bids.SessionConverter(
            session_id="456",
            nwbfile_paths=[minimal_nwbfile_path],
            session_metadata=nwb2bids.bids_models.BidsSessionMetadata(
                session_id="456",
                participant=nwb2bids.bids_models.Participant(
                    participant_id="123", species="Mus musculus", sex="M", strain=None
                ),
            ),
        ),
    ]
    assert dataset_converter.session_converters == expected_session_converters


def test_dataset_converter_write_dataset_description(
    minimal_nwbfile_path: pathlib.Path,
    additional_metadata_file_path: pathlib.Path,
    temporary_bids_directory: pathlib.Path,
):
    nwb_paths = [minimal_nwbfile_path]
    dataset_converter = nwb2bids.DatasetConverter.from_nwb_paths(
        nwb_paths=nwb_paths, additional_metadata_file_path=additional_metadata_file_path
    )
    dataset_converter.extract_metadata()
    dataset_converter.write_dataset_description(bids_directory=temporary_bids_directory)

    expected_structure = {temporary_bids_directory: {"directories": set(), "files": {"dataset_description.json"}}}
    nwb2bids.testing.assert_subdirectory_structure(
        directory=temporary_bids_directory, expected_structure=expected_structure
    )

    dataset_description_file_path = temporary_bids_directory / "dataset_description.json"
    with dataset_description_file_path.open(mode="r") as file_stream:
        dataset_description_json = json.load(fp=file_stream)

    expected_dataset_description = {
        "Name": "test",
        "Description": "TODO",
        "BIDSVersion": "1.10",
        "DatasetType": "raw",
        "License": "CC-BY-4.0",
        "Authors": ["Cody Baker", "Yaroslav Halchenko"],
    }
    assert dataset_description_json == expected_dataset_description


def test_dataset_converter_write_subject_metadata(
    minimal_nwbfile_path: pathlib.Path, temporary_bids_directory: pathlib.Path
):
    nwb_paths = [minimal_nwbfile_path]
    dataset_converter = nwb2bids.DatasetConverter.from_nwb_paths(nwb_paths=nwb_paths)
    dataset_converter.extract_metadata()

    dataset_converter.write_participants_metadata(bids_directory=temporary_bids_directory)

    expected_structure = {
        temporary_bids_directory: {
            "directories": set(),
            "files": {"dataset_description.json", "participants.json", "participants.tsv"},
        }
    }
    nwb2bids.testing.assert_subdirectory_structure(
        directory=temporary_bids_directory, expected_structure=expected_structure
    )

    participants_tsv_file_path = temporary_bids_directory / "participants.tsv"
    participants_data_frame = pandas.read_csv(filepath_or_buffer=participants_tsv_file_path, sep="\t", dtype=str)

    expected_data_frame = pandas.DataFrame({"participant_id": ["sub-123"], "species": ["Mus musculus"], "sex": ["M"]})
    pandas.testing.assert_frame_equal(left=participants_data_frame, right=expected_data_frame)

    participants_json_file_path = temporary_bids_directory / "participants.json"
    with participants_json_file_path.open(mode="r") as file_stream:
        participants_json = json.load(fp=file_stream)
    expected_participants_json = {
        "participant_id": "A unique identifier for this participant.",
        "species": (
            "The species should be the proper Latin binomial species name from the NCBI Taxonomy "
            "(for example, Mus musculus)."
        ),
        "sex": (
            'String value indicating phenotypical sex, one of "male", "female", "other".\n'
            '\tFor "male", use one of these values: male, m, M, MALE, Male.\n'
            '\tFor "female", use one of these values: female, f, F, FEMALE, Female.\n'
            '\tFor "other", use one of these values: other, o, O, OTHER, Other.'
        ),
    }
    assert participants_json == expected_participants_json


def test_dataset_converter_write_sessions_metadata(
    minimal_nwbfile_path: pathlib.Path, temporary_bids_directory: pathlib.Path
):
    nwb_paths = [minimal_nwbfile_path]
    dataset_converter = nwb2bids.DatasetConverter.from_nwb_paths(nwb_paths=nwb_paths)
    dataset_converter.extract_metadata()

    dataset_converter.write_sessions_metadata(bids_directory=temporary_bids_directory)

    expected_structure = {
        temporary_bids_directory: {"directories": {"sub-123"}, "files": {"dataset_description.json"}},
        temporary_bids_directory
        / "sub-123": {
            "directories": {"ses-456"},
            "files": {"sub-123_sessions.json", "sub-123_sessions.tsv"},
        },
        temporary_bids_directory
        / "sub-123"
        / "ses-456": {
            "directories": set(),
            "files": set(),
        },
    }
    nwb2bids.testing.assert_subdirectory_structure(
        directory=temporary_bids_directory, expected_structure=expected_structure
    )

    sessions_tsv_file_path = temporary_bids_directory / "sub-123" / "sub-123_sessions.tsv"
    sessions_data_frame = pandas.read_csv(filepath_or_buffer=sessions_tsv_file_path, sep="\t", dtype=str)

    expected_data_frame = pandas.DataFrame({"session_id": ["ses-456"]})
    pandas.testing.assert_frame_equal(left=sessions_data_frame, right=expected_data_frame)

    sessions_json_file_path = temporary_bids_directory / "sub-123" / "sub-123_sessions.json"
    with sessions_json_file_path.open(mode="r") as file_stream:
        sessions_json = json.load(fp=file_stream)
    expected_sessions_json = {"session_id": "A unique session identifier."}
    assert sessions_json == expected_sessions_json



import pathlib
import re

import pydantic
import pynwb
import typing_extensions

from ._model_globals import _VALID_ARCHIVES_SEXES, _VALID_BIDS_SEXES, _VALID_ID_REGEX, _VALID_SPECIES_REGEX
from .._inspection._inspection_result import Category, DataStandard, InspectionResult, Severity
from ..bids_models._base_metadata_model import BaseMetadataModel


class Participant(BaseMetadataModel):
    participant_id: str = pydantic.Field(
        description="A unique identifier for this participant.",
        default=None,
    )
    species: str | None = pydantic.Field(
        description=(
            "The species should be the proper Latin binomial species name from the NCBI Taxonomy "
            "(for example, Mus musculus)."
        ),
        default=None,
    )
    sex: str | None = pydantic.Field(
        description=(
            'String value indicating phenotypical sex, one of "male", "female", "other".\n'
            '\tFor "male", use one of these values: male, m, M, MALE, Male.\n'
            '\tFor "female", use one of these values: female, f, F, FEMALE, Female.\n'
            '\tFor "other", use one of these values: other, o, O, OTHER, Other.'
        ),
        default=None,
    )
    strain: str | None = pydantic.Field(
        description=(
            "For species other than Homo sapiens, the string value indicating the strain of the species "
            "(for example, C57BL/6J)."
        ),
        default=None,
    )

    def _check_fields(self, file_paths: list[pathlib.Path] | list[pydantic.HttpUrl]) -> None:
        # Check if values are specified
        if self.participant_id is None:
            self.messages.append(
                InspectionResult(
                    title="Missing participant ID",
                    reason="A unique ID is required for all individual participants.",
                    solution="Specify the `subject_id` field of the Subject object attached to the NWB file.",
                    field="nwbfile.subject.subject_id",
                    source_file_paths=file_paths,
                    data_standards=[DataStandard.BIDS, DataStandard.DANDI],
                    category=Category.SCHEMA_INVALIDATION,
                    severity=Severity.CRITICAL,
                )
            )
        if self.species is None:
            self.messages.append(
                InspectionResult(
                    title="Missing participant species",
                    reason="Archives such as DANDI or EMBER require the subject species to be specified.",
                    solution=(
                        "Specify the `species` field of the Subject object attached to the NWB file as a "
                        "Latin binomial, obolibrary taxonomy link, or NCBI taxonomy reference."
                    ),
                    field="nwbfile.subject.species",
                    source_file_paths=file_paths,
                    data_standards=[DataStandard.DANDI],
                    category=Category.SCHEMA_INVALIDATION,
                    severity=Severity.CRITICAL,
                )
            )
        if self.sex is None:
            self.messages.append(
                InspectionResult(
                    title="Missing participant sex",
                    reason="Archives such as DANDI or EMBER require the subject sex to be specified.",
                    solution=(
                        "Specify the `sex` field of the Subject object attached to the NWB file as one of four "
                        'options: "M" (for male), "F" (for female), "U" (for unknown), or "O" (for other).\n'
                        "NOTE: for certain animal species with more specific genetic determinants, such as C elegans, "
                        'use "O" (for other) then further specify the subtypes using other custom fields. For example, '
                        '`c_elegans_sex="XO"`'
                    ),
                    field="nwbfile.subject.sex",
                    source_file_paths=file_paths,
                    data_standards=[DataStandard.DANDI],
                    category=Category.SCHEMA_INVALIDATION,
                    severity=Severity.CRITICAL,
                )
            )

        # Check if specified values are valid
        if self.participant_id is not None and re.match(pattern=_VALID_ID_REGEX, string=self.participant_id) is None:
            self.messages.append(
                InspectionResult(
                    title="Invalid participant ID",
                    reason=(
                        "The participant ID contains invalid characters. "
                        "BIDS allows only dashes to be used as separators in subject entity label. "
                        "Underscores, spaces, slashes, and special characters (including #) are expressly forbidden."
                    ),
                    solution="Rename the subject without using spaces or underscores.",
                    examples=[
                        "`ab_01` -> `ab-01`",
                        "`subject #2` -> `subject-2`",
                        "`id 2 from 9/1/25` -> `id-2-9-1-25`",
                    ],
                    field="nwbfile.subject.subject_id",
                    source_file_paths=file_paths,
                    data_standards=[DataStandard.BIDS, DataStandard.DANDI],
                    category=Category.STYLE_SUGGESTION,
                    severity=Severity.ERROR,
                )
            )
        if self.species is not None and re.match(pattern=_VALID_SPECIES_REGEX, string=self.species) is None:
            self.messages.append(
                InspectionResult(
                    title="Invalid species",
                    reason="Participant species is not a proper Latin binomial or NCBI Taxonomy id.",
                    solution=(
                        "Rename the subject species to a Latin binomial, obolibrary taxonomy link, or "
                        "NCBI taxonomy reference."
                    ),
                    examples=[],
                    field="nwbfile.subject.species",
                    source_file_paths=file_paths,
                    data_standards=[DataStandard.DANDI],
                    category=Category.SCHEMA_INVALIDATION,
                    severity=Severity.ERROR,
                )
            )
        if self.sex is not None and _VALID_BIDS_SEXES.get(self.sex, None) is None:
            self.messages.append(
                InspectionResult(
                    title="Invalid participant sex (BIDS)",
                    reason="Participant sex is not one of the allowed patterns by BIDS.",
                    solution="Rename the subject sex to be one of the accepted values.",
                    examples=["`male` -> `M`", "`Female` -> `F`", "`n/a` -> `U`", "`hermaphrodite` -> `O`"],
                    field="nwbfile.subject.sex",
                    source_file_paths=file_paths,
                    data_standards=[DataStandard.BIDS],
                    category=Category.STYLE_SUGGESTION,
                    severity=Severity.ERROR,
                )
            )

        if self.sex is not None and _VALID_ARCHIVES_SEXES.get(self.sex, None) is None:
            self.messages.append(
                InspectionResult(
                    title="Invalid participant sex (archives)",
                    reason="Participant sex is not one of the allowed patterns by the common archives.",
                    solution="Rename the subject sex to be one of the accepted values.",
                    examples=["`male` -> `M`", "`Female` -> `F`", "`n/a` -> `U`", "`hermaphrodite` -> `O`"],
                    field="nwbfile.subject.sex",
                    source_file_paths=file_paths,
                    data_standards=[DataStandard.DANDI],
                    category=Category.STYLE_SUGGESTION,
                    severity=Severity.ERROR,
                )
            )

    @classmethod
    @pydantic.validate_call
    def from_nwbfiles(cls, nwbfiles: list[pydantic.InstanceOf[pynwb.NWBFile]]) -> typing_extensions.Self:
        """
        Extracts participant metadata from the in-memory NWBFile objects.
        """
        file_paths = [nwbfile.container_source for nwbfile in nwbfiles]

        messages = []
        if len(nwbfiles) > 1:
            messages.append(
                InspectionResult(
                    title="NotImplemented: multiple NWB files",
                    reason=(
                        "The `Participant` model for `nwb2bids` does not yet support multiple NWB files. "
                        "Only the first will be used."
                    ),
                    solution="`nwb2bids` plans to add support for multiple NWB files in the future.",
                    source_file_paths=file_paths,
                    category=Category.INTERNAL_ERROR,
                    severity=Severity.ERROR,
                )
            )

        nwbfile = nwbfiles[0]

        if nwbfile.subject is None:
            messages.append(
                InspectionResult(
                    title="Missing participant",
                    reason="BIDS requires a subject to be specified for each NWB file.",
                    solution="Add a Subject object to each NWB file.",
                    field="nwbfile.subject",
                    source_file_paths=file_paths,
                    data_standards=[DataStandard.BIDS, DataStandard.DANDI],
                    category=Category.STYLE_SUGGESTION,
                    severity=Severity.CRITICAL,
                )
            )
            participant = cls(messages=messages)
            return participant

        participant = cls(
            participant_id=nwbfile.subject.subject_id,
            species=nwbfile.subject.species,
            sex=nwbfile.subject.sex,
            strain=nwbfile.subject.strain,
            messages=messages,
            # TODO: add more
            # birthday=nwbfile.participant.date_of_birth,
            # age=nwbfile.participant.age
        )
        participant._check_fields(file_paths=file_paths)
        return participant



"""
Integration tests for edge cases of the primary `convert_nwb_dataset` function.

Namely, the cases of:
  - A dataset with additional metadata and the case of a dataset.
  - A dataset without a session ID.
"""

import pathlib

import nwb2bids


def test_convert_nwb_dataset_with_additional_metadata(
    minimal_nwbfile_path: pathlib.Path,
    temporary_bids_directory: pathlib.Path,
    additional_metadata_file_path: pathlib.Path,
):
    nwb_paths = [minimal_nwbfile_path]
    nwb2bids.convert_nwb_dataset(
        nwb_paths=nwb_paths,
        bids_directory=temporary_bids_directory,
        additional_metadata_file_path=additional_metadata_file_path,
    )

    expected_structure = {
        temporary_bids_directory: {
            "directories": {"sub-123"},
            "files": {"participants.json", "participants.tsv", "dataset_description.json"},
        },
        temporary_bids_directory
        / "sub-123": {
            "directories": {"ses-456"},
            "files": {"sub-123_sessions.json", "sub-123_sessions.tsv"},
        },
        temporary_bids_directory
        / "sub-123"
        / "ses-456": {
            "directories": {"ecephys"},
            "files": set(),
        },
        temporary_bids_directory
        / "sub-123"
        / "ses-456"
        / "ecephys": {
            "directories": set(),
            "files": {"sub-123_ses-456_ecephys.nwb"},
        },
    }
    nwb2bids.testing.assert_subdirectory_structure(
        directory=temporary_bids_directory, expected_structure=expected_structure
    )



import enum
import json
import pathlib

import pydantic


@enum.unique
class DataStandard(enum.Enum):
    """Related standards used by the `nwb2bids` inspections."""

    DANDI = enum.auto()
    HED = enum.auto()
    NWB = enum.auto()
    BIDS = enum.auto()


@enum.unique
class Category(enum.Enum):
    """Types of inspection categories."""

    STYLE_SUGGESTION = enum.auto()
    SCHEMA_INVALIDATION = enum.auto()
    INTERNAL_ERROR = enum.auto()


@enum.unique
class Severity(enum.Enum):
    """
    Quantifier of relative severity (how important it is to resolve) for inspection results.

    The larger the value, the more critical it is.

    If an issue can be categorized in multiple ways, the most severe category should be chosen.
    """

    INFO = enum.auto()  # Not an indication of problem but information of status or confirmation
    HINT = enum.auto()  # Data is valid but could be improved
    WARNING = enum.auto()  # Data is not recognized as valid. Changes are needed to ensure validity
    ERROR = enum.auto()  # Data is recognized as invalid
    CRITICAL = enum.auto()  # A serious invalidity in data


class InspectionResult(pydantic.BaseModel):
    title: str = pydantic.Field(description="Short title of the issue.")
    reason: str = pydantic.Field(description="Detailed description of the issue and suggestions for how to resolve it.")
    solution: str = pydantic.Field(
        description="Detailed description of the issue and suggestions for how to resolve it."
    )
    examples: list[str] | None = pydantic.Field(
        description="Detailed description of the issue and suggestions for how to resolve it.", default=None
    )
    field: str | None = pydantic.Field(
        description=(
            "Best description of the location within the file where the issue was detected "
            "(such as field name, group, etc.)."
        ),
        default=None,
    )
    source_file_paths: list[pathlib.Path] | list[pydantic.HttpUrl] | None = pydantic.Field(
        description="If known, the paths of all source NWB file(s) where the issue was detected.",
        default=None,
    )
    target_file_paths: list[pathlib.Path] | list[pydantic.HttpUrl] | None = pydantic.Field(
        description=(
            "If known, the target BIDS paths of all file(s) associated with the session where the issue was detected."
        ),
        default=None,
    )
    data_standards: list[DataStandard] | None = pydantic.Field(
        description="Data standard(s) related to the issue.", default=None
    )
    category: Category = pydantic.Field(description="Type of inspection category.")
    severity: Severity = pydantic.Field(
        description="Quantifier of relative severity. The larger the value, the more important it is.",
    )

    # TODO: remove this when/if PyNWB fixes source container for remfile objects
    @pydantic.field_validator("source_file_paths", mode="after")
    def validate_source_file_paths(cls, value: list[str] | None) -> list[str] | None:
        """Remove any paths that contain 'remfile', which NWB source container fields include automatically."""
        if value is None:
            return None
        scrubbed = [path for path in value if "remfile" not in str(path)]
        scrubbed_source_file_paths = None if len(scrubbed) == 0 else scrubbed
        return scrubbed_source_file_paths

    def model_dump(self, **kwargs) -> dict:
        mode = kwargs.pop("mode", "python")
        python_data = super().model_dump(mode="python", **kwargs)

        if mode == "json":
            json_data = json.loads(_CustomJSONEncoder().encode(python_data))
            return json_data

        return python_data


class _CustomJSONEncoder(json.JSONEncoder):
    """
    Required to generate custom desired JSON output when using the Enum fields.

    Without the enum fields,
    """

    def default(self, obj):
        if isinstance(obj, enum.Enum):
            return obj.name
        elif isinstance(obj, pydantic.HttpUrl):
            return str(obj)
        elif isinstance(obj, pathlib.Path):
            return str(obj)
        return super().default(obj)



import abc
import json
import pathlib
import tempfile
import typing

import pydantic


class BaseConverter(pydantic.BaseModel, abc.ABC):
    @abc.abstractmethod
    def extract_metadata(self) -> None:
        """
        Extract essential metadata used by the BIDS standard from the source NWB files.
        """
        message = f"The `extract_metadata` method has not been implemented by the `{self.__class__.__name__}` class."
        raise NotImplementedError(message)

    def _handle_bids_directory(self, bids_directory: str | pathlib.Path | None = None) -> pathlib.Path:
        """
        Handle the BIDS directory path.

        If the directory does not exist, create it.
        If it exists, validate that it is a valid BIDS dataset.
        """
        if bids_directory is None:
            bids_directory = pathlib.Path.cwd()
        bids_directory = pathlib.Path(bids_directory)

        if bids_directory.exists():
            self._validate_existing_directory_as_bids(bids_directory=bids_directory)
        else:
            bids_directory.mkdir(exist_ok=True)
        self.extract_metadata()

        return bids_directory

    @staticmethod
    def _validate_existing_directory_as_bids(bids_directory: pathlib.Path) -> None:
        """
        Validate that the existing directory is a valid BIDS dataset.

        Parameters
        ----------
        bids_directory : pathlib.Path
            The path to the directory to validate.
        """
        dataset_description_file_path = bids_directory / "dataset_description.json"

        current_directory_contents = {
            path.stem for path in bids_directory.iterdir() if not path.name.startswith(".")
        } - {"README", "CHANGES", "derivatives", "dandiset"}
        if len(current_directory_contents) == 0:
            default_dataset_description = {"BIDSVersion": "1.10"}
            with dataset_description_file_path.open(mode="w") as file_stream:
                json.dump(obj=default_dataset_description, fp=file_stream, indent=4)
            return

        if not dataset_description_file_path.exists():
            message = (
                f"The directory ({bids_directory}) exists and is not empty, but is not a valid BIDS dataset: "
                "missing 'dataset_description.json'."
            )
            raise ValueError(message)

        with dataset_description_file_path.open(mode="r") as file_stream:
            dataset_description = json.load(fp=file_stream)
        if dataset_description.get("BIDSVersion", None) is None:
            message = (
                f"The directory ({bids_directory}) exists but is not a valid BIDS dataset: "
                "missing 'BIDSVersion' in 'dataset_description.json'."
            )
            raise ValueError(message)

    @staticmethod
    def _handle_file_mode(
        file_mode: typing.Literal["move", "copy", "symlink", "auto"] = "auto",
    ) -> typing.Literal["move", "copy", "symlink"]:
        if file_mode != "auto":
            return file_mode

        with tempfile.TemporaryDirectory(prefix="nwb2bids-") as temp_dir_str:
            temp_dir_path = pathlib.Path(temp_dir_str)

            # Create a test file
            test_file_path = temp_dir_path / "test_file.txt"
            test_file_path.touch()

            try:
                # Create a symlink to the test file
                (temp_dir_path / "test_symlink.txt").symlink_to(target=test_file_path)
            except (OSError, PermissionError, NotImplementedError):  # Windows can sometimes have trouble with symlinks
                # TODO: log a INFO message here when logging is set up
                return "copy"
            else:
                # If symlink creation was successful, return "symlink"
                return "symlink"



from ._bids_session_metadata import BidsSessionMetadata, Participant
from ._dataset_description import DatasetDescription
from ._probes import ProbeTable, Probe
from ._electrodes import Electrode, ElectrodeTable
from ._channels import Channel, ChannelTable

__all__ = [
    "BidsSessionMetadata",
    "Channel",
    "ChannelTable",
    "DatasetDescription",
    "Electrode",
    "ElectrodeTable",
    "Participant",
    "Probe" "ProbeTable",
]



_VALID_ID_REGEX = r"^[A-Za-z0-9]+$"
_VALID_SPECIES_REGEX = r"([A-Z][a-z]* [a-z]+)|(http://purl.obolibrary.org/obo/NCBITaxon_\d+)"
_VALID_BIDS_SEXES = {
    value: True
    for value in [
        "male",
        "m",
        "M",
        "MALE",
        "Male",
        "female",
        "f",
        "F",
        "FEMALE",
        "Female",
        "other",
        "o",
        "O",
        "OTHER",
        "Other",
        "n/a",
    ]
}
_VALID_ARCHIVES_SEXES = {
    value: True
    for value in [
        "M",
        "F",
        "O",
        "U",
    ]
}

_BIDS_RRID = "RRID:SCR_016124"



import json
import pathlib

import pandas
import pydantic
import pynwb
import typing_extensions

from .._inspection._inspection_result import InspectionResult
from ..bids_models._base_metadata_model import BaseMetadataContainerModel, BaseMetadataModel


class Probe(BaseMetadataModel):
    probe_id: str
    type: str | None = None
    description: str | None = None
    manufacturer: str | None = None


class ProbeTable(BaseMetadataContainerModel):
    probes: list[Probe]

    @pydantic.computed_field
    @property
    def messages(self) -> list[InspectionResult]:
        """
        All messages from contained session converters.

        These can accumulate over time based on which instance methods have been called.
        """
        messages = [message for probe in self.probes for message in probe.messages]
        messages.sort(key=lambda message: (-message.category.value, -message.severity.value, message.title))
        return messages

    @classmethod
    @pydantic.validate_call
    def from_nwbfiles(cls, nwbfiles: list[pydantic.InstanceOf[pynwb.NWBFile]]) -> typing_extensions.Self | None:
        electrical_series = [
            neurodata_object
            for nwbfile in nwbfiles
            for neurodata_object in nwbfile.objects.values()
            if isinstance(neurodata_object, pynwb.ecephys.ElectricalSeries)
        ]
        if any(electrical_series) is False:
            return None

        unique_devices = {
            electrode_group.device
            for nwbfile in nwbfiles
            for electrode_group in nwbfile.electrodes["group"][:]
            if nwbfile.electrodes is not None
        }

        probes = [
            Probe(
                probe_id=device.name,
                type=None,  # TODO
                description=device.description,
                manufacturer=device.manufacturer,
            )
            for device in unique_devices
        ]
        return cls(probes=probes)

    @pydantic.validate_call
    def to_tsv(self, file_path: str | pathlib.Path) -> None:
        """
        Save the probes information to a TSV file.

        Parameters
        ----------
        file_path : path
            The path to the output TSV file.
        """
        data = []
        for probe in self.probes:
            model_dump = probe.model_dump()
            data.append(model_dump)

        data_frame = pandas.DataFrame(data=data)
        data_frame.to_csv(path_or_buf=file_path, sep="\t", index=False)

    @pydantic.validate_call
    def to_json(self, file_path: str | pathlib.Path) -> None:
        """
        Save the probes information to a JSON file.

        Parameters
        ----------
        file_path : path
            The path to the output JSON file.
        """
        with file_path.open(mode="w") as file_stream:
            json.dump(obj=dict(), fp=file_stream, indent=4)



from ._assert_subdirectory_structure import assert_subdirectory_structure
from ._mocks._mock_neurodata_objects import mock_time_intervals, mock_epochs_table, mock_trials_table

__all__ = [
    "assert_subdirectory_structure",
    "mock_epochs_table",
    "mock_time_intervals",
    "mock_trials_table",
]



import json
import pathlib

import pandas
import pydantic
import pynwb
import typing_extensions

from .._inspection._inspection_result import InspectionResult
from ..bids_models._base_metadata_model import BaseMetadataContainerModel, BaseMetadataModel


class Electrode(BaseMetadataModel):
    electrode_id: int
    probe_id: str
    location: str | None = None


class ElectrodeTable(BaseMetadataContainerModel):
    electrodes: list[Electrode]

    @pydantic.computed_field
    @property
    def messages(self) -> list[InspectionResult]:
        """
        All messages from contained session converters.

        These can accumulate over time based on which instance methods have been called.
        """
        messages = [message for electrode in self.electrodes for message in electrode.messages]
        messages.sort(key=lambda message: (-message.category.value, -message.severity.value, message.title))
        return messages

    @classmethod
    @pydantic.validate_call
    def from_nwbfiles(cls, nwbfiles: list[pydantic.InstanceOf[pynwb.NWBFile]]) -> typing_extensions.Self | None:
        if len(nwbfiles) > 1:
            message = "Conversion of multiple NWB files per session is not yet supported."
            raise NotImplementedError(message)
        nwbfile = nwbfiles[0]

        electrical_series = [
            neurodata_object
            for neurodata_object in nwbfile.objects.values()
            if isinstance(neurodata_object, pynwb.ecephys.ElectricalSeries)
        ]
        if any(electrical_series) is False:
            return None

        electrodes = [
            Electrode(
                electrode_id=electrode.index[0],
                probe_id=electrode.group.iloc[0].device.name,
                # TODO "impedance": electrode["imp"].iloc[0] if electrode["imp"].iloc[0] > 0 else None,
                location=str(electrode.location.values[0]),
            )
            for electrode in nwbfile.electrodes
        ]
        return cls(electrodes=electrodes)

    @pydantic.validate_call
    def to_tsv(self, file_path: str | pathlib.Path) -> None:
        """
        Write the electrode data to a TSV file.

        Parameters
        ----------
        file_path : path
            The path to the output TSV file.
        """
        data = []
        for electrode in self.electrodes:
            model_dump = electrode.model_dump()
            data.append(model_dump)

        data_frame = pandas.DataFrame(data=data)
        data_frame.to_csv(file_path, sep="\t", index=False)

    @pydantic.validate_call
    def to_json(self, file_path: str | pathlib.Path) -> None:
        """
        Save the electrode information to a JSON file.

        Parameters
        ----------
        file_path : path
            The path to the output JSON file.
        """
        with file_path.open(mode="w") as file_stream:
            json.dump(obj=dict(), fp=file_stream, indent=4)



import pynwb


def mock_trials_table() -> pynwb.epoch.TimeIntervals:
    """
    Generate a mock trials table.

    These are shorter than epochs and fit neatly within them in a qualitative manner.
    """
    trials = pynwb.epoch.TimeIntervals(name="trials", description="A mock trials table.")
    trials.add_column(name="trial_condition", description="Extra information per trial.")
    trials.add_row(start_time=0.0, stop_time=1.0, trial_condition="A")
    trials.add_row(start_time=2.0, stop_time=3.0, trial_condition="B")
    trials.add_row(start_time=5.0, stop_time=5.5, trial_condition="C")
    trials.add_row(start_time=5.5, stop_time=6.0, trial_condition="D")

    return trials


def mock_epochs_table() -> pynwb.epoch.TimeIntervals:
    """
    Generate a mock epochs table.

    These are longer than trials and encompass several of them in a qualitative manner.
    """
    epochs = pynwb.epoch.TimeIntervals(name="epochs", description="A mock epochs table.")
    epochs.add_column(name="epoch_condition", description="Extra information per epoch.")
    epochs.add_row(start_time=0.0, stop_time=3.5, epoch_condition="AB")
    epochs.add_row(start_time=4.5, stop_time=6.0, epoch_condition="CD")

    return epochs


def mock_time_intervals() -> pynwb.epoch.TimeIntervals:
    """Generate a mock time intervals table."""
    time_intervals = pynwb.epoch.TimeIntervals(name="mock_time_intervals", description="A mock time intervals table.")
    time_intervals.add_column(name="tag", description="A tag assigned to each interval.")
    time_intervals.add_row(start_time=0.5, stop_time=0.75, tag="tag1")
    time_intervals.add_row(start_time=5.1, stop_time=5.23, tag="tag2")

    return time_intervals
